/**
 * @file Attribute.cc
 *
 *  @remark ~ Implementation of the Attribute class.  
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  . THIS software was developed at the National Institute of           .
 *  . Standards and Technology by employees of the Federal Government    .
 *  . in the course of their official duties. Pursuant to title 17       .
 *  . Section 105 of the United States Code this software is not         .
 *  . subject to copyright protection and is in the public               .
 *  . domain. NIST's [intelligent mobility] software is an experimental  .
 *  . system. NIST assumes no responsibility whatsoever for its use by   .
 *  . other parties, and makes no guarantees, expressed or implied,      .
 *  . about its quality, reliability, or any other characteristic. We    .
 *  . would appreciate acknowledgement if the software is used. This     .
 *  . software can be redistributed and/or modified freely provided      .
 *  . that any derivative works bear some notice that they are derived   .
 *  . from it, and any modified versions bear some notice that they      .
 *  . have been modified.                                                .
 *   `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-01-17 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "ViewController_Attrib.hh"

#ifdef _WIN32
#  define WIN32_LEAN_AND_MEAN 1
#endif

/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;


/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[ViewController_Attrib %d]: ",     \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#ifdef NDEBUG
#define fakeUseParam
#else
static void fakeUseParam(...) {}
#endif

AttributeClass::
AttributeClass ()
{
  m_keys_hashP = g_hash_table_new (g_str_hash, g_str_equal);
}

inline static int 
CompString (const void *key1, 
            const void *key2, 
            const void *context)
{
  fakeUseParam (context);
  return strcmp ((const char*) key1, (const char*) key2);
}


AttributeClass::
AttributeClass (AttributeType *in_attributeA,
                int            in_numKeys)
{
  m_keys_hashP = g_hash_table_new (g_str_hash, g_str_equal);
  for (int idx = 0; idx < in_numKeys; idx++)
    {
//       g_hash_table_insert  (m_keys_hashP, 
//                             in_attributeA [idx].name, 
//                             strdup (in_attributeA [idx].value));
      g_hash_table_insert  (m_keys_hashP, 
                            strdup (in_attributeA [idx].name), 
                            strdup (in_attributeA [idx].value));
    }

}



static gboolean
FreeKeyAndValue (gpointer key,
                 gpointer value,
                 gpointer )
{
  g_free (key);
  g_free (value);

  return true;
}



AttributeClass::
~AttributeClass ()
{
  Reset ();
  g_hash_table_destroy (m_keys_hashP);
  m_keys_hashP = NULL;
}


void AttributeClass::
Reset ()
{
  g_hash_table_foreach_remove (m_keys_hashP, FreeKeyAndValue, NULL);
}


void AttributeClass::
SetValue (const char *in_key, 
          const char *in_value)
{
//   fprintf (stderr, "calling SetValue, in_key = %s\n", in_key);

  gpointer value =  g_hash_table_lookup (m_keys_hashP, in_key);

  /* create an entry if key is not found: */
  if (value == NULL)
    {
      value = strdup (in_value);
      g_hash_table_insert (m_keys_hashP, strdup (in_key), value);
    }

  /* replace with the new value: */
  else
    {
      free (value);
      value = strdup (in_value);
      g_hash_table_insert (m_keys_hashP, (gpointer) in_key, value);
    }

  assert (g_hash_table_lookup (m_keys_hashP, in_key));
//   fprintf (stderr, "Set: key = '%s', value = '%s'\n",
//            (char*) in_key, (char*) value);
}


const char* AttributeClass::
GetValue (const char *in_key)
{
  gpointer value =  g_hash_table_lookup (m_keys_hashP, in_key);
  if (value == NULL)
    {
      fprintf (stderr, "no value for key '%s'\n", in_key);
      assert (value);
    }
  if (strcmp ((const char*) value, "") == 0)
    return NULL;

  return (const char*) value;
}



void AttributeClass::
AppendValue (const char *in_key, 
             const char *in_value)
{
  fakeUseParam (in_value, in_key);

  gpointer value =  g_hash_table_lookup (m_keys_hashP, in_key);

  /* no pre-existing value: */
  if (value == NULL)
    {
      SetValue (in_key, in_value);
      return;
    }
    
  /* append to the end of pre-existing value: */
  g_hash_table_insert (m_keys_hashP, (gpointer) in_key, 
                       g_strdup_printf ("%s %s", (char*) value, in_value));
  free (value);
}


bool AttributeClass::
CheckKey (const char *in_key)
{
  gpointer value =  g_hash_table_lookup (m_keys_hashP, in_key);
  if (value == NULL)
    return false;

  return true;
}



bool AttributeClass::
LoadConfig (const char *in_fileName_str)
{
//    = m_attributeObjP->GetValue ("configFile");
//   assert (in_fileName_str);

  FILE *fileP = fopen (in_fileName_str, "rb");
  if (fileP == NULL)
    {
      fprintf (stderr, "warning, could not load config file %s\n",
               in_fileName_str);
      return false;
    }

  
  const int maxLen = 512;       // 512 should be enough for everybody
  char line_str [maxLen];

  while (likely (fgets (line_str, maxLen, fileP) != NULL))
    {
      assert (strlen (line_str) < (maxLen - 1)); // check memory.
      if (line_str [0] == '\n') // skip blank lines
        continue;
      if ((line_str [0] == '\r') && (line_str [1] == '\n'))// skip blank lines
        continue;
      if (line_str [0] == '#') // skip comment lines
        continue;


      /* get property and its value : */
//       char *propertyName_str = NULL;
//       char *propertyVal_str = NULL;
//       int retVal = sscanf (line_str, "%a[^=]%a[^\n]", // name=val
//                            &propertyName_str, &propertyVal_str);
//       assert (retVal == 2);

//       if (strlen (propertyVal_str) > 1) // skip the '=' sign
//         {
//           SetValue (propertyName_str, propertyVal_str + 1);
//           fprintf (stderr, "property name:%s, value:%s\n",
//                    propertyName_str, propertyVal_str + 1);
//         }
//       free (propertyName_str);
//       free (propertyVal_str);
      char propertyName_str[512]; // 512 should be enough for everybody
      char propertyVal_str[512];  // 512 should be enough for everybody
      int retVal = sscanf (line_str, "%[^=]%[^\n]", // name=val
                           propertyName_str, propertyVal_str);
      assert (retVal == 2);

      /* check dos format: */
      if (propertyVal_str [strlen (propertyVal_str) - 1] == '\r')
        propertyVal_str [strlen (propertyVal_str) - 1] = 0;

      if (strlen (propertyVal_str) > 1) // skip the '=' sign
        {
          SetValue (propertyName_str, propertyVal_str + 1);
          fprintf (stderr, "property name:%s, value:%s\n",
                   propertyName_str, propertyVal_str + 1);
        }

    }
  
  fclose (fileP);
  return true;
}

