/**
 * @file Attribute.hh
 *
 *  @remark ~ The attribute class used by the Controller and View modules.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  . THIS software was developed at the National Institute of           .
 *  . Standards and Technology by employees of the Federal Government    .
 *  . in the course of their official duties. Pursuant to title 17       .
 *  . Section 105 of the United States Code this software is not         .
 *  . subject to copyright protection and is in the public               .
 *  . domain. NIST's [intelligent mobility] software is an experimental  .
 *  . system. NIST assumes no responsibility whatsoever for its use by   .
 *  . other parties, and makes no guarantees, expressed or implied,      .
 *  . about its quality, reliability, or any other characteristic. We    .
 *  . would appreciate acknowledgement if the software is used. This     .
 *  . software can be redistributed and/or modified freely provided      .
 *  . that any derivative works bear some notice that they are derived   .
 *  . from it, and any modified versions bear some notice that they      .
 *  . have been modified.                                                .
 *   `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-01-17 
 **/


#ifndef ATTRIBUTE_HH
#define ATTRIBUTE_HH
/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <gtk/gtk.h>


/*-----------------*
 * Macro Constants *
 \*---------------*/



/*-----------------*
 * Data Structures *
 \*---------------*/
typedef struct {
  const char *name;
  const char *value;
} AttributeType;

class AttributeClass
{
public:
  AttributeClass ();
  AttributeClass (AttributeType *in_attributeA,
                  int            in_numKeys);
  ~AttributeClass ();

  /** <!--SetValue()-->
   * @brief ~ Set the value of the attribute
   *
   *  @param[in] in_key ~ the key of the attribute
   *  @param[in] in_value ~ the value to be associated with the key;
   *
   *  @retval none
   *
   *  @remark ~ Call this function to set the key-value look-up.
   *
   *  @pre ~ none
   *  @post ~ pre-existing value is overwritten; A copy of *in_value is made
   *      so it   doesn't matter if *in_value is later freed or not.
   *
   *  @see ~ AppendValue()
   **/
  void        SetValue (const char *in_key, const char *in_value);


  const char* GetValue (const char *in_key);
  bool        CheckKey (const char *in_key);
  bool        LoadConfig (const char *in_fileName_str);

  /** <!--AppendValue()-->
   * @brief ~ Append to the value
   *
   *  @param[in] in_key ~ the key
   *  @param[in] in_value ~ the value to append
   *
   *  @retval none 
   *
   *  @remark ~ Call this function to append value to the end of the existing
   *      value.
   *
   *  @pre ~ none
   *  @post ~ A copy of *in_value is made so it's ok that *in_value changes
   *      later.
   *
   *  @see ~ SetValue()
   **/
  void AppendValue (const char *in_key, const char *in_value);

  void Reset ();

private:
  GHashTable *m_keys_hashP;
};



/*--------------------*
 * Function Prototype *
 \*------------------*/


#endif // ATTRIBUTE_HH


