/**
 * @file gCardSys.cc
 *
 *  @remark ~ Front-end Implementation of the Leitner System.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright 2009 Tommy Chang.  This file is part of gCardSys.
 *
 *  gCardSys is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  gCardSys is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gCardSys.  If not, see <http://www.gnu.org/licenses/>.
 *  `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-05-31 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <getopt.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <libgen.h>             // need by dirname

#include <gtk/gtk.h>
#include "gCardSys_Controller.hh"

typedef struct {
  int         debugLevel_bitored;
  const char *statusFile_str;
  const char *guiFile_str;
  const char *title_str;
  int         exportDkf_int;
} CommandLineParamType;



/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;


#ifdef IS_HILDON
#  include <hildon/hildon-program.h>
#  include <libosso.h>
 /* Build up the D-Bus name for this application (NEW). */
#  define PACKAGE_DBUS_NAME "org.maemo." PACKAGE_NAME
#  define BUILDERXML "gCardSys_2-10.ui"
#else
#  define BUILDERXML "gCardSys.ui"
#endif

static CommandLineParamType g_clParams = {
  0,                            // debugLevel_bitored
  NULL,                         // statusFile_str
  BUILDERXML,                   // guiFile_str, memory leak, oh well
  "gCardSys",                   // title_str
  0,                            // exportDkf_int
};

static AttributeType g_attributeA[]  = {
  {"key", "value"},
  {"exportDkf", "0"},
};

static AttributeClass g_ctrlAttributeObj 
(g_attributeA, sizeof (g_attributeA) / sizeof (AttributeType)); 


/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[gCardSys %d]: ",                  \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#ifdef NDEBUG
#define fakeUseParam(...)
#else
static void fakeUseParam(...) {}
#endif



static void
ExitFailure (const char *mesg_str = NULL)
{
  fakeUseParam ();
  if (mesg_str)
    fprintf (stderr, "%s\n", mesg_str);

  if (g_clParams.statusFile_str)
    {
      FILE *sFileP = fopen (g_clParams.statusFile_str, "wb");
      if (sFileP)
        {
          fprintf (sFileP, "0"); // write 0 to status file -- failure
          fclose (sFileP);
        }
      else
        fprintf (stderr, "could not open file %s\n", 
                 g_clParams.statusFile_str);


      if (mesg_str)
        {
          char *mesgFileName = (char*)
            malloc (strlen (g_clParams.statusFile_str) + 10);
          sprintf (mesgFileName, "%s.msg", g_clParams.statusFile_str);
          fprintf (stderr, "writing file to %s\n", mesgFileName);
          FILE *sFileP = fopen (mesgFileName, "wb");
          if (sFileP)
            {
              fprintf (sFileP, "%s", mesg_str); 
              fclose (sFileP);
            }
          else
            fprintf (stderr, "could not open file %s\n", mesgFileName);
        }
    }
  exit (EXIT_FAILURE);
}



static void
CheckRequiredOptions ()
{
  

}


static void
SetSystemReady ()
{
  // system ready:
  if (g_clParams.statusFile_str)
    {
      FILE *sFileP = fopen (g_clParams.statusFile_str, "wb");
      fprintf (sFileP, "%d", getpid ()); // success = pid to status file 
      fclose (sFileP);
    }
}



static bool
InitSystem ()
{
  static LeitnerModelClass  modelObj;
  static ViewClass          viewObj (&modelObj);
  static ControllerClass    controllerObj (&viewObj, &g_ctrlAttributeObj);

  /* load all the gui widgets: */
  GtkBuilder *builder = gtk_builder_new ();
  if (gtk_builder_add_from_file (builder, g_clParams.guiFile_str, NULL) == 0)
    {
      fprintf (stderr, "gui error, check file '%s;\n", g_clParams.guiFile_str);
      return false;
    }
  gtk_builder_connect_signals (builder, &controllerObj); 

  /* initialize the MVC system :*/
  controllerObj.Init (builder, g_clParams.title_str);

  return true;
}


static void
Run ()
{
  /*--------------------------------------
   * Check Require Commandline options:
   *------------------------------------*/
  CheckRequiredOptions ();

  /*-----------------------------
   * subsystems initialization:
   *---------------------------*/
  if (InitSystem ())
    SetSystemReady ();
  else
    ExitFailure ("InitSystem failed");

  /*-----------------------------
   * controller loop:
   *---------------------------*/
  fprintf (stderr, "Calling main loop\n");
  gtk_main ();
}



static void
PrintUsage (char *progname_param)
{
  char *progname = strrchr (progname_param, '\\');
  if (progname == NULL)
    progname = progname_param;
  else
    progname += 1;

  fprintf 
    (stderr,
     "\n\nLast compiles on: %s\n"
     "Usage: %s [options]:\n"
     "[options] =\n"
     "  --help                Show this usage instruction.\n"
     "  --debug <int>         <int> is the level of verboseness;\n" 
     "                            can be multi-valued; default 0.\n" 
     "  --statusFile <str>    <str> is the output status file.\n" 
     "                            The status file contains the pid \n" 
     "                            if success, otherwise 0.\n" 
     "  --                    Allow no option supplied -- no effect.\n"
     "  --title <str>         <str> is the window title to use\n"
     "  --guiFile <str>       <str> is the gui file to use\n"
     "                            (default gCardSys.xml)\n"
     "  --exportDkf           Export all cards to a single dkf when saving\n"
     "\n"
     "examples:\n"
     "  %s --\n"
     "  %s --debug 1 --debug 3\n"
     "\n", __DATE__
     , progname
     , progname
     , progname);
  exit (0);
}



static void
ParseCmdLine (int    argc,
              char **argv)
{
  /*---------------------------------
   * be intelligent about arguments
   *-------------------------------*/
  static struct option longOpt[] = {
    /*name, has_arg, int* flg, if flg: *flg = retval, ow. get_long() = retval*/
    {"debug",           1, NULL,                      0},
    {"help",            0, NULL,                     'h'},
    {"statusFile",      1, NULL,                      0},
    {"title",           1, NULL,                      0},
    {"guiFile",         1, NULL,                      0},
    {"exportDkf",       0, NULL,                      0},
    {0,0,0,0} // terminator
  };


//   if (argc == 1)
//     PrintUsage (argv[0]);

  int opt = 0;
  int optIdx = 0;
# define IsOpt_M(pname) else if (strcmp (longOpt [optIdx].name, pname) == 0)
  while ((opt = getopt_long_only (argc, argv, "", longOpt, &optIdx)) != -1)
    {
      switch (opt)              // optarg is the argument
        {
        case 0:
          if (longOpt [optIdx].flag) // auto assigned value
            break;
          IsOpt_M ("debug")
            g_clParams.debugLevel_bitored |= (1 << (atoi (optarg) - 1));
          IsOpt_M ("statusFile")
            g_clParams.statusFile_str = optarg;
          IsOpt_M ("guiFile")
	    g_clParams.guiFile_str = optarg;
          IsOpt_M ("title")
	    g_clParams.title_str = optarg;
          IsOpt_M ("exportDkf")
            g_ctrlAttributeObj.SetValue ("exportDkf", "1");
          else 
            fprintf (stderr, "un-handled option %s\n", longOpt [optIdx].name);
          break;
        case 'h':               // get help
          PrintUsage (argv[0]);
          exit (EXIT_SUCCESS);
          break;
        case '?':               // unrecognized option, no more msg needed
          ExitFailure ();
          break;
        default:
          printf ("%c option not implemented yet\n", opt);
          break;
        }
    }
# undef IsOpt_M

  if (optind < argc)
    {
      for (; optind < argc; optind++)
        fprintf (stderr, "unknown extra parameter %s\n", argv[optind]);
      ExitFailure ();
    }


  /*
   * set the gui file to bin location if not specified:
   */
  if (strcmp (g_clParams.guiFile_str, BUILDERXML) == 0)
    {
      fprintf (stderr, "executable is %s\n", argv[0]);
      char *exeCmpPath_str = strdup (argv [0]);  // never freed

      char *exePath_str = dirname (exeCmpPath_str);
      char *guiCmpPath_str = (char*) 
        malloc (strlen (exePath_str) + strlen (BUILDERXML) + 10);
      
      sprintf (guiCmpPath_str, "%s/%s", exePath_str, BUILDERXML);
      g_clParams.guiFile_str = guiCmpPath_str;
      fprintf (stderr, "gui file changed to %s\n", g_clParams.guiFile_str);
    }
}


int
main (int    argc,
      char **argv)
{
  /*------------------------------
   * parse command-line arguments:
   *----------------------------*/
  ParseCmdLine (argc, argv);
  gtk_init (&argc, &argv);      // initialize GTK stuff

#ifdef NO_HILDON_DBUS
#else
  /* 
   * Create a LibOSSO context (which will also attach this
   * application to the D-Bus (NEW).
   * NOTE:
   *   We use the name and version from the configure.ac.
   *   The D-Bus name is built by prefixing "org.maemo." to the
   *   package name. 
   */
  fprintf (stderr, "Initializing LibOSSO context (" PACKAGE_DBUS_NAME ", "
           PACKAGE_VERSION ")\n");
  osso_context_t* ctx = 
    osso_initialize (PACKAGE_DBUS_NAME, PACKAGE_VERSION, TRUE, NULL);
  if (ctx == NULL) 
    {
      fprintf (stderr, "Failed to init LibOSSO\n");
      return EXIT_FAILURE;
    }
  fprintf (stderr, "LibOSSO Init done\n");
#endif


  /*-----------------------------
   * run system:
   *---------------------------*/
  Run ();


  /*-----------------------------
   * program termination:
   *---------------------------*/
#ifdef NO_HILDON_DBUS
#else
  fprintf (stderr, "main: returned from gtk_main & de-initing LibOSSO\n");
  /* De-initialize LibOSSO (detaches from the D-Bus)*/
  osso_deinitialize(ctx);
#endif

  return (EXIT_SUCCESS);
}
