/**
 * @file flashCard.cc
 *
 *  @remark ~ Implementation of flash card supported functions.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright (c) 2009 Tommy Chang. Permission is granted to copy,       .
 *  distribute and/or modify this document under the terms of the GNU    .
 *  Free Documentation License, Version 1.3 or any later version         .
 *  published by the Free Software Foundation; with no Invariant         .
 *  Sections, no Front-Cover Texts, and no Back-Cover Texts. A copy of   .
 *  the license is included in the file fdl-1.3.txt.                     .
 *  `--------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-06-07 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>             // need by get_current_dir_name
#include <libgen.h>             // need by dirname

#include "flashCard.hh"

/* TC 2010-03-28 (Sun) -- turn off debugging */
#ifdef NDEBUG_FLASHCARD
#define NDEBUG
#endif

/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;


/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[flashCard %d]: ",                 \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#ifdef NDEBUG
#define fakeUseParam(...)
#else
static void fakeUseParam(...) {}
#endif


static void 
FreeFlashCard (FlashCardType *flashCardP)
{
//   fprintf (stderr, "flashCardP = %p\n", flashCardP);
//   fprintf (stderr, "front_str = '%s'\n", flashCardP->front_str);
//   fprintf (stderr, "back_str = '%s'\n", flashCardP->back_str);
//   fprintf (stderr, "backExample_str = '%p'\n", flashCardP->backExample_str);
//   fprintf (stderr, "backExample_str = '%s'\n", flashCardP->backExample_str);

  free (flashCardP->front_str); 
  free (flashCardP->back_str); 
  free (flashCardP->backExample_str);
  free (flashCardP->backPicFile_str);
  free (flashCardP->frontPicFile_str);
  free (flashCardP);
}


inline static int 
CompTsDictNode (const void *key1, 
                const void *key2, 
                const void *,
                const void *,
                const void *)
{
  if ((uint32_t) key1 < (uint32_t) key2)
    return -1;

  if ((uint32_t) key1 > (uint32_t) key2)
    return 1;

  return 0;
}




FlashCardClass::
FlashCardClass () :
  m_loaded_flg (false),
  m_deckXmlState_enum (DECK_XML_STATE_NONE),
  m_tmpFlashCardP (NULL),
  m_flashCardsPA (NULL),
  m_dictP (NULL),
  m_soundPath_str (NULL),
  m_fullSoundPath_str (NULL),
  m_picPath_str (NULL),
  m_fullDeckPath_str (NULL)
{
//   fprintf (stderr, "calling constructor\n");
  // m_loaded_flg = false;
  // m_deckXmlState_enum = DECK_XML_STATE_NONE;

  // m_flashCardsPA = NULL;
  // m_tmpFlashCardP = NULL;

  m_dictP = dict_create (DICTCOUNT_T_MAX, CompTsDictNode, NULL); 
  // m_soundPath_str = NULL;
  // m_picPath_str = NULL;
}


FlashCardClass::
~FlashCardClass ()
{
//   fprintf (stderr, "calling destructor\n");
  Unload ();

  dict_destroy (m_dictP);
  m_dictP = NULL;

  if (m_flashCardsPA)
    {
      free (m_flashCardsPA);
      m_flashCardsPA = NULL;
    }
}

static void
AddToDict (dict_t        *in_dictP,
           FlashCardType *in_flashCardP)
{
  /*
   * Create a dictionary node and add data to the dictionary:
   */
  dnode_t *dnodeP = dnode_create (in_flashCardP);
  dict_insert (in_dictP, dnodeP, (void*) (in_flashCardP->wordUiq_id));
}


void FlashCardClass::
ProcessXmlNode (xmlTextReaderPtr in_readerP)
{
  /*
   * Get the name and value:
   */
  xmlChar *elmName = xmlTextReaderName (in_readerP);
  xmlChar *elmValue = xmlTextReaderValue (in_readerP);
  xmlChar *attName, *attValue;

  /* debugging: */
//   fprintf (stderr, "elmName is '%s'\n", elmName);
//   fprintf (stderr, "elmValue is '%s'\n", elmValue);
  // fprintf (stderr, "type is %d\n",
  //          xmlTextReaderNodeType (in_readerP));

  switch (m_deckXmlState_enum) 
    {
    case DECK_XML_STATE_NONE:
      /*
       * Look for "sound_path" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "sound_path") == 0))
        {
          m_deckXmlState_enum = DECK_XML_STATE_SOUND_PATH;
          if (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              /* get the "relative" attribute: */
              xmlTextReaderMoveToNextAttribute (in_readerP);
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

              assert (strcmp ((char*) attName, "relative") == 0);
              // fprintf (stderr, "attValue is %s\n", attValue);

              // TC 2010-05-29 (Sat) -- not doing anything with the
              // relative attribute.  TBF

              /* clean ups: */
              xmlFree (attName);
              xmlFree (attValue);
            }
        }
      break;

    case DECK_XML_STATE_SOUND_PATH:
      /*
       * Look for "#text" node:
       */
      /* no optional sound_path: */
      if (xmlTextReaderNodeType (in_readerP) == 14)
        {
          if (elmValue [0] == '\n') // will it work in Dos format?
            m_deckXmlState_enum = DECK_XML_STATE_SOUND_PATH_TEXT;
        }

      /* get the optional sound_path: */
      if (xmlTextReaderNodeType (in_readerP) == 3)
        {
          m_deckXmlState_enum = DECK_XML_STATE_SOUND_PATH_TEXT;

          fprintf (stderr, "sound path is '%s'\n", elmValue);
          assert (m_soundPath_str == NULL);
          m_soundPath_str = strdup ((char*) elmValue);
        }
      break;

    case DECK_XML_STATE_SOUND_PATH_TEXT:
    case DECK_XML_STATE_PICS_PATH_TEXT: // same state here
      /*
       * Look for "card" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "card") == 0))
        {
          m_deckXmlState_enum = DECK_XML_STATE_CARD;

          /* get the "id" attribute: */
          xmlTextReaderMoveToNextAttribute (in_readerP);
          attName = xmlTextReaderName (in_readerP);
          attValue = xmlTextReaderValue (in_readerP);

          assert (strcmp ((char*) attName, "id") == 0);

          /* allocate storage and insert into the list: */
          m_tmpFlashCardP = (FlashCardType*) calloc (1, sizeof (FlashCardType));
          m_tmpFlashCardP->wordUiq_id =  atoi ((char*) attValue + 1); // skip _
          AddToDict (m_dictP, m_tmpFlashCardP);

          /* clean ups: */
          xmlFree (attName);
          xmlFree (attValue);
        }
      else
        /*
         * Look for "pics_path" start element: 
         * (Although there can be more than one pics_path, I just store
         * the last one. TC 2009-08-30 --TBF)
         */
        if ((xmlTextReaderNodeType (in_readerP) == 1) &&
            (strcmp ((char*) elmName, "pics_path") == 0))
          {
            m_deckXmlState_enum = DECK_XML_STATE_PICS_PATH;

          if (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              /* get the "relative" attribute: */
              xmlTextReaderMoveToNextAttribute (in_readerP);
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

              assert (strcmp ((char*) attName, "relative") == 0);
              // fprintf (stderr, "attValue is %s\n", attValue);

              // TC 2010-05-29 (Sat) -- not doing anything with the
              // relative attribute.  TBF

              /* clean ups: */
              xmlFree (attName);
              xmlFree (attValue);
            }
          }
      
      break;

    case DECK_XML_STATE_PICS_PATH:
      /*
       * Look for "EMPTY" node:
       */
      /* Empty: */
      if (xmlTextReaderNodeType (in_readerP) == 14)
        {
          if (elmValue [0] == '\n') // will it work in Dos format?
            m_deckXmlState_enum = DECK_XML_STATE_PICS_PATH_TEXT;
        }

      // The granted.dtd says EMPTY for pics path
      // /* get the optional pics_path: */
      // if (xmlTextReaderNodeType (in_readerP) == 3)
      //   {
      //     m_deckXmlState_enum = DECK_XML_STATE_PICS_PATH_TEXT;

      //     fprintf (stderr, "pics path is '%s'\n", elmValue);
      //     assert (m_picPath_str == NULL);
      //     m_picPath_str = strdup ((char*) elmValue);
      //   }
      break;

    case DECK_XML_STATE_CARD:
      /*
       * Look for "front" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "front") == 0))
        {
          m_deckXmlState_enum = DECK_XML_STATE_FRONT;

          if (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              /* get the "image" attribute: */
              xmlTextReaderMoveToNextAttribute (in_readerP);
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

              assert (strcmp ((char*) attName, "image") == 0);
              // fprintf (stderr, "attValue is %s\n", attValue);

              /* allocate storage and fill the image field: */
              m_tmpFlashCardP->frontPicFile_str = strdup ((char*) attValue);

              /* clean ups: */
              xmlFree (attName);
              xmlFree (attValue);
            }
        }
      break;

    case DECK_XML_STATE_FRONT:
      /*
       * Look for "#text" node:
       */
      if (xmlTextReaderNodeType (in_readerP) == 3)
        {
          m_deckXmlState_enum = DECK_XML_STATE_FRONT_TEXT;

          /* allocate storage and fill the front_str: */
          m_tmpFlashCardP->front_str = strdup ((char*) elmValue);
        }
      break;

    case DECK_XML_STATE_FRONT_TEXT:
      /*
       * Look for "back" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "back") == 0))
        m_deckXmlState_enum = DECK_XML_STATE_BACK;
      break;

    case DECK_XML_STATE_BACK:
      /*
       * Look for "#text" node:
       */
      if (xmlTextReaderNodeType (in_readerP) == 3)
        {
          m_deckXmlState_enum = DECK_XML_STATE_BACK_TEXT;

          /* allocate storage and insert into the list: */
          m_tmpFlashCardP->back_str = strdup ((char*) elmValue);
        }

      /* Check for empty: */
      if ((m_tmpFlashCardP->back_str == NULL) &&
          (xmlTextReaderNodeType (in_readerP) == 15)) // end of element
        {
          m_deckXmlState_enum = DECK_XML_STATE_BACK_TEXT;
          m_tmpFlashCardP->back_str = strdup ("");
          // fprintf (stderr, "******* empty back_str\n");
        }


      break;

    case DECK_XML_STATE_BACK_TEXT:
      /*
       * Look for "back_example" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "back_example") == 0))
        {
          m_deckXmlState_enum = DECK_XML_STATE_BACK_EXAMPLE;

          if (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              /* get the "image" attribute: */
              xmlTextReaderMoveToNextAttribute (in_readerP);
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

              assert (strcmp ((char*) attName, "image") == 0);
              // fprintf (stderr, "attValue is %s\n", attValue);

              /* allocate storage and fill the image field: */
              m_tmpFlashCardP->backPicFile_str = strdup ((char*) attValue);

              /* clean ups: */
              xmlFree (attName);
              xmlFree (attValue);
            }
        }
      break;
      
    case DECK_XML_STATE_BACK_EXAMPLE:
      /*
       * Look for "#text" node:
       */
      if (xmlTextReaderNodeType (in_readerP) == 3)
        {
          m_deckXmlState_enum = DECK_XML_STATE_SOUND_PATH_TEXT;

          /* allocate storage and insert into the list: */
          m_tmpFlashCardP->backExample_str = strdup ((char*) elmValue);
          // fprintf (stderr, "backExample_str is '%s'\n", 
          //          m_tmpFlashCardP->backExample_str);
          // const xmlChar *xml_str = xmlEncodeEntitiesReentrant 
          //   (NULL, (xmlChar*) m_tmpFlashCardP->backExample_str);
          // fprintf (stderr, "backExample_str is '%s'\n", (char*) xml_str);
        }

      /* Check for empty: */
      if ((m_tmpFlashCardP->backExample_str == NULL) &&
          (xmlTextReaderNodeType (in_readerP) == 15)) // end of element
        {
          m_deckXmlState_enum = DECK_XML_STATE_SOUND_PATH_TEXT;
          m_tmpFlashCardP->backExample_str = strdup ("");
          // fprintf (stderr, "******* empty backexample_str\n");
        }
      break;
    default:
      break;
    }

  /*
   * clean ups:
   */
  xmlFree (elmName);
  if (elmValue)
    xmlFree (elmValue);
}


bool FlashCardClass::
Load (const char *in_deckFile_str,
      int        *out_numCards_intP)
{
  fprintf (stderr, "==> loading deck from %s\n", in_deckFile_str);

  /*
   * Make sure not loaded already:
   */
  if (out_numCards_intP)
    *out_numCards_intP = 0; // if failed, we set out_numCards_intP anyway

  if (m_loaded_flg)
    return false;


  /*
   * Open the deck file for XML parsing:
   */
  xmlTextReaderPtr readerP = xmlNewTextReaderFilename (in_deckFile_str);
  if (readerP == NULL)
    {
      fprintf (stderr, "could not read file %s\n", in_deckFile_str);
      return false;
    }
  

  /*
   * Parse the deck file and fill data:
   */
  m_deckXmlState_enum = DECK_XML_STATE_NONE;
  int ret = xmlTextReaderRead (readerP);
  while (ret == 1)
    {
      ProcessXmlNode (readerP);
      ret = xmlTextReaderRead (readerP); // XML doc can not have un-escaped &
    }

  /* Check XML data integrity: */
  assert (ret == 0);
  if (ret != 0)
    {
      fprintf (stderr, "parser error!!!\n");
      return false;
    }

  /* 
   * clean up (file closed): 
   */
  xmlTextReaderClose (readerP);
  xmlFreeTextReader (readerP);


  /*
   * data is now loaded:
   */
  fprintf (stderr, "loaded %d cards\n", (int) dict_count (m_dictP));
  if (out_numCards_intP)
    *out_numCards_intP = (int) dict_count (m_dictP);
  m_loaded_flg = true;


  /*
   * Set the full deck and sound directories:
   */
  char *fullCurDir_str = get_current_dir_name (); 
  char *deckFile_str = strdup (in_deckFile_str);
  char *deckDir = strdup (dirname (deckFile_str));
  free (deckFile_str);
  // fprintf (stderr, "deckDir is %s\n", deckDir);

  // temporaray change to the deck dir to get its full path:
  chdir (deckDir);
  free (deckDir);
  assert (m_fullDeckPath_str == NULL);
  m_fullDeckPath_str = get_current_dir_name (); // allocated here

  // temporary change to the sound dir to get its full path:
  if (m_soundPath_str)
    {
      chdir (m_soundPath_str);
      // fprintf (stderr, "m_soundPath_str is '%s'\n", m_soundPath_str);
      assert (m_fullSoundPath_str == NULL);
      m_fullSoundPath_str = get_current_dir_name (); // allocated here
    }

  // restore to original dir:
  chdir (fullCurDir_str);
  free (fullCurDir_str);


  /*
   * Set the full sound directory:
   */


  return true;
}

// static void
// RemoveFromDict (dnode_t        *in_dnodeP,
// {
//   /*
//    * Create a dictionary node and add data to the dictionary:
//    */
//   dnode_t *dnodeP = dnode_create (in_flashCardP);
//   dict_insert (in_dictP, dnodeP, (void*) (in_flashCardP->wordUiq_id));
// }



void FlashCardClass::
Unload ()
{
  if (m_loaded_flg)
    {
      /*
       * Free all cards from the dictionary:
       */
      dnode_t *dnodeP = dict_first (m_dictP);
      while (likely (dnodeP))
        {
          FlashCardType *flashCardP = (FlashCardType*) dnode_get (dnodeP);
          assert (flashCardP);
          FreeFlashCard (flashCardP);

          dnode_t *free_dnodeP = dnodeP;
          dnodeP = dnodeP->next;
          dict_delete_free (m_dictP, free_dnodeP); 
          // TC 2009-06-11 --free the dnode, may not be the efficient,
          // when loading again, consider using memory pool
        }
      assert (dict_count (m_dictP) == 0);

      /*
       * free the paths:
       */
      if (m_soundPath_str)
        {
          free (m_soundPath_str);
          m_soundPath_str = NULL;
        }
      if (m_fullSoundPath_str)
        {
          free (m_fullSoundPath_str);
          m_fullSoundPath_str = NULL;
        }
      // if (m_picPath_str)
      //   {
      //     free (m_picPath_str);
      //     m_picPath_str = NULL;
      //   }
      if (m_fullDeckPath_str)
	{
	  free (m_fullDeckPath_str);
	  m_fullDeckPath_str = NULL;
	}

      fprintf (stderr, "cards unloaded\n");
    }

  m_loaded_flg = false;
  m_deckXmlState_enum = DECK_XML_STATE_NONE;
}



const FlashCardType** FlashCardClass::
GetAllCards (int *out_numCards_intP)
{
  if (out_numCards_intP)
    *out_numCards_intP = 0;     // in case failed, we set output anyway.

  if (m_loaded_flg == false)
    return NULL;

  /*
   * Re-Create a "serialized" array:
   */
  m_flashCardsPA = (FlashCardType**) 
    realloc (m_flashCardsPA, dict_count (m_dictP) * sizeof (FlashCardType*));
  assert (m_flashCardsPA);

  /*
   * put data onto the "serialized" array:
   */
  dnode_t *dnodeP = dict_first (m_dictP);
  // fprintf (stderr, "there are %d in m_dictP\n", (int) dict_count (m_dictP));
  int cardIdx = 0;
  while (likely (dnodeP))
    {
      FlashCardType *flashCardP = (FlashCardType*) dnode_get (dnodeP);
      assert (flashCardP);
      m_flashCardsPA [cardIdx++] = flashCardP;
      dnodeP = dnodeP->next;
    };


  /*
   * output the number of cards:
   */
  if (out_numCards_intP)
    *out_numCards_intP = dict_count (m_dictP);

  
  return (const FlashCardType**) m_flashCardsPA;
}



const FlashCardType* FlashCardClass::
GetCard (uint32_t in_card_id)
{
  if (m_loaded_flg == false)
    return NULL;

  dnode_t *dnodeP = dict_lookup (m_dictP, (void*) in_card_id, NULL);
  if (dnodeP == NULL)
    return NULL;

  return (FlashCardType*) dnode_get (dnodeP);
}


bool FlashCardClass::
ChangeFields (uint32_t     in_wordUiq_id,
              const char  *in_front_str,
              const char  *in_back_str,
              const char  *in_backExample_str,
              const char  *in_frontPicFile_str,
              const char  *in_backPicFile_str)
{
  if (m_loaded_flg == false)
    return false;
  
  FlashCardType* flashCardP = (FlashCardType*) GetCard (in_wordUiq_id);
  if (flashCardP == NULL)
    return false;

  if (in_front_str)
    {
      free (flashCardP->front_str); 
      flashCardP->front_str = strdup (in_front_str);
    }

  if (in_back_str)
    {
      free (flashCardP->back_str);
      flashCardP->back_str = strdup (in_back_str);
    }

  if (in_backExample_str)
    {
      free (flashCardP->backExample_str);
      flashCardP->backExample_str = strdup (in_backExample_str);
    }
  
  if (in_backPicFile_str)
    {
      free (flashCardP->backPicFile_str);
      flashCardP->backPicFile_str = strdup (in_backPicFile_str);
    }
  
  if (in_frontPicFile_str)
    {
      free (flashCardP->frontPicFile_str);
      flashCardP->frontPicFile_str = strdup (in_frontPicFile_str);
    }

  return true;
}


static const char*
ConvertToXmlFormatStr (const char *in_str)
{
  fakeUseParam ();
  return (const char*) xmlEncodeEntitiesReentrant (NULL, (xmlChar*) in_str);
}

static void
WriteDeckHeader (FILE       *in_fileP,
                 const char *in_soundPath_str)
{
  fprintf (in_fileP, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  fprintf (in_fileP, "<!DOCTYPE deck SYSTEM \"granule.dtd\">\n");
  fprintf (in_fileP, "<deck>\n");
  fprintf (in_fileP, "  <author>Tommy Chang</author>\n");
  fprintf (in_fileP, "  <description>Created by Tommy Chang</description>\n");
  fprintf (in_fileP, "  <sound_path relative=\"yes\">%s</sound_path>\n",
           in_soundPath_str);
  fprintf (in_fileP, "  <pics_path relative=\"yes\"/>\n");
  fprintf (in_fileP, "  <appearance enabled=\"no\"/>\n");
}

static void
WriteCardToFile (FILE                *in_fileP,
                 const FlashCardType *in_flashCardP)
{
  fprintf (in_fileP, "  <card id=\"_%010d\">\n", 
           in_flashCardP->wordUiq_id);

  /*
   * front_str: 
   */
  if (in_flashCardP->frontPicFile_str)
    fprintf (in_fileP, "    <front image=\"%s\">%s</front>\n", 
             ConvertToXmlFormatStr (in_flashCardP->frontPicFile_str),
             ConvertToXmlFormatStr (in_flashCardP->front_str));
  else
    fprintf (in_fileP, "    <front>%s</front>\n", 
             ConvertToXmlFormatStr (in_flashCardP->front_str));

  /*
   * back_str: 
   */
  fprintf (in_fileP, "    <back>%s</back>\n", 
           ConvertToXmlFormatStr (in_flashCardP->back_str));
  
  /*
   * back_example: 
   */
  if (in_flashCardP->backPicFile_str)
    fprintf (in_fileP, "    <back_example image=\"%s\">", // no \n
             ConvertToXmlFormatStr (in_flashCardP->backPicFile_str));
  else
    fprintf (in_fileP, "    <back_example>"); // no "\n"
  fprintf (in_fileP, "%s", 
           ConvertToXmlFormatStr (in_flashCardP->backExample_str));
  fprintf (in_fileP, "</back_example>\n"); // note: there is no indent
  fprintf (in_fileP, "  </card>\n");
}


bool FlashCardClass::
Save (const char *in_saveFile_str)
{
  if (m_loaded_flg == false)
    {
      fprintf (stderr, "nothing to save\n");
      return true;
    }

  /*
   * Open file for writing:
   */
  FILE *fileP = fopen (in_saveFile_str, "wb");
  if (fileP == NULL)
    {
      fprintf (stderr, "could not open file for saving\n");
      return false;
    }

  /*
   * Write the file header block:
   */
  // fprintf (fileP, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  // fprintf (fileP, "<!DOCTYPE deck SYSTEM \"granule.dtd\">\n");
  // fprintf (fileP, "<deck>\n");
  // fprintf (fileP, "  <author>Tommy Chang</author>\n");
  // fprintf (fileP, "  <description>Created by Tommy Chang</description>\n");
  // fprintf (fileP, "  <sound_path relative=\"yes\">%s</sound_path>\n",
  //          m_soundPath_str);
  // fprintf (fileP, "  <pics_path relative=\"yes\"/>\n");
  // fprintf (fileP, "  <appearance enabled=\"no\"/>\n");
  WriteDeckHeader (fileP, m_soundPath_str);


  int numCards;
  const FlashCardType **flashCardsPA = GetAllCards (&numCards);

  for (int idx = 0; idx < numCards; idx++)
    {
      // fprintf (fileP, "  <card id=\"_%010d\">\n", 
      //          flashCardsPA [idx]->wordUiq_id);
      // if (flashCardsPA [idx]->frontPicFile_str)
      //   fprintf (fileP, "    <front image=\"%s\">%s</front>\n", 
      //            ConvertToXmlFormatStr (flashCardsPA [idx]->frontPicFile_str),
      //            ConvertToXmlFormatStr (flashCardsPA [idx]->front_str));
      // else
      //   fprintf (fileP, "    <front>%s</front>\n", 
      //            ConvertToXmlFormatStr (flashCardsPA [idx]->front_str));
      // fprintf (fileP, "    <back>%s</back>\n", 
      //          ConvertToXmlFormatStr (flashCardsPA [idx]->back_str));

      // if (flashCardsPA [idx]->backPicFile_str)
      //   fprintf (fileP, "    <back_example image=\"%s\">", // no \n
      //            ConvertToXmlFormatStr (flashCardsPA [idx]->backPicFile_str));
      // else
      //   fprintf (fileP, "    <back_example>"); // no "\n"
      // fprintf (fileP, "%s", 
      //          ConvertToXmlFormatStr (flashCardsPA [idx]->backExample_str));
      // fprintf (fileP, "</back_example>\n"); // note: there is no indent
      // fprintf (fileP, "  </card>\n");

      WriteCardToFile (fileP, flashCardsPA [idx]);
    }


  /*
   * Close file after writing:
   */
  fprintf (fileP, "</deck>\n");
  fclose (fileP);

  return true;
}



void FlashCardClass::
PrintCardInfo (const FlashCardType *in_cardP)
{
  fprintf (stderr, "----- CARD INFO -----\n");
  fprintf (stderr, "front_str = '%s'\n", in_cardP->front_str);
  fprintf (stderr, "frontPicFile_str = '%s'\n", in_cardP->frontPicFile_str);
  fprintf (stderr, "----- CARD INFO -----\n\n");
}




//=====================
// DeckSavingClass
//=====================
inline static int 
CompFlashCardWordDictNode (const void *key1, // word1
                           const void *key2, // word2
                           const void *,
                           const void *,
                           const void *)
{
  return strcmp ((char*) key1, (char*) key2);
}

DeckSavingClass::
DeckSavingClass (const char *in_deckFile_str,
                 const char *in_soundPath_str) :
  m_dictP (NULL),
  m_fileP (NULL),
  m_soundPath_str (NULL)
{
  m_fileP = fopen (in_deckFile_str, "wb");
  m_dictP = dict_create (DICTCOUNT_T_MAX, CompFlashCardWordDictNode, NULL); 
  dict_allow_dupes (m_dictP);
  m_soundPath_str = strdup (in_soundPath_str);
}


bool DeckSavingClass::
InsertCard (const FlashCardType *in_flashCardP)
{
  FlashCardType *flashCardP = new FlashCardType;
  memset (flashCardP, 0, sizeof (FlashCardType));
  assert (in_flashCardP);

  // flashCardP->wordUiq_id = 0; // computed at the saving time
  if (in_flashCardP->front_str) // strdup doesn't dup NULL very well...
    flashCardP->front_str = strdup (in_flashCardP->front_str);
  if (in_flashCardP->back_str)
    flashCardP->back_str = strdup (in_flashCardP->back_str);
  if (in_flashCardP->backExample_str)
    flashCardP->backExample_str = strdup (in_flashCardP->backExample_str);
  if (in_flashCardP->frontPicFile_str)
    flashCardP->frontPicFile_str = strdup (in_flashCardP->frontPicFile_str);
  if (in_flashCardP->backPicFile_str)
    flashCardP->backPicFile_str = strdup (in_flashCardP->backPicFile_str);

  dnode_t *dnodeP = dnode_create (flashCardP);
  dict_insert (m_dictP, dnodeP, (void*) (flashCardP->front_str));
  return true;
}

DeckSavingClass::
~DeckSavingClass ()
{
  assert (m_fileP);
  if (m_fileP == NULL)
    {
      FreeMemory ();
      return;
    }

  /*
   * Save to new deck file:
   */

  /* Write the file header block: */
  WriteDeckHeader (m_fileP, m_soundPath_str);

  /* Write the word block: */
  dnode_t *dnodeP = dict_first (m_dictP);
  int wordUiq_id = 0;
  while (likely (dnodeP))
    {
      FlashCardType *flashCardP = (FlashCardType*) dnode_get (dnodeP);
      flashCardP->wordUiq_id = wordUiq_id++;
      WriteCardToFile (m_fileP, flashCardP);
      dnodeP = dnodeP->next;
    }


  /*
   * Close file and clean up:
   */
  fprintf (m_fileP, "</deck>\n");
  fclose (m_fileP);
  FreeMemory ();
}


void DeckSavingClass::
FreeMemory ()
{
  free (m_soundPath_str);

  /*
   * Free memory:
   */
  dnode_t *dnodeP = dict_first (m_dictP);
  while (likely (dnodeP))
    {
      FlashCardType *flashCardP = (FlashCardType*) dnode_get (dnodeP);
      assert (flashCardP);
      FreeFlashCard (flashCardP);
      
      dnode_t *free_dnodeP = dnodeP;
      dnodeP = dnodeP->next;
      dict_delete_free (m_dictP, free_dnodeP); 
    }
  assert (dict_count (m_dictP) == 0);
  dict_destroy (m_dictP);
}
