/**
 * @file leitnerSysModel.cc
 *
 *  @remark ~ The back-end implementation of the Leitner system.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright 2009 Tommy Chang.  This file is part of gCardSys.
 *
 *  gCardSys is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  gCardSys is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gCardSys.  If not, see <http://www.gnu.org/licenses/>.
 *  `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-05-31 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>               // needed by clock_gettime()
#include <unistd.h>             // getpid
#include <libgen.h>             // need by dirname

#include "leitnerSysModel.hh"


/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;
#define DEFAULT_PAUSE_TIMEOUT_MIN 4 // default value for pause timeout

/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[leitnerSysModel %d]: ",           \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#ifdef NDEBUG
#define fakeUseParam(...)
#else
static void fakeUseParam(...) {}
#endif


#define PoolSize(var) *((unsigned int*) ((int*) (var) - 1))

#define Free(var) do {                          \
  if (var)                                      \
    free ((((int*) (var)) - 1));                \
  var = NULL;                                   \
} while (0)

#define Malloc(var, type, size) do {                            \
    if ((var) && ((unsigned int) (size) <= PoolSize (var)))     \
      break;                                                    \
    if (var)                                                    \
      Free ((var));                                             \
    var = (type) (((int*) malloc (sizeof (int) + (size))) + 1); \
    PoolSize (var) = size;                                      \
  } while (0)


#define UseUserStorage(out_AP, type, size)      \
  do {                                          \
    if (out_AP && *out_AP)                      \
      Malloc (*out_AP, type, size);             \
  } while (0)


#define UseBuiltInStorage(out_AP, builtInA)     \
  do {                                          \
    if (out_AP && (*out_AP == NULL))            \
      *out_AP = builtInA;                       \
  } while (0)


static uint32_t
GetMinSinceEpoch ()
{
  struct timespec the_timeSpec;
  clock_gettime (CLOCK_REALTIME, &the_timeSpec);
  return the_timeSpec.tv_sec / 60; // 1 min = 60 sec
}

static void 
FreeDeckFile (xmlLinkPtr lk)
{
  xmlFree (xmlLinkGetData (lk));
}

static void 
FreeCardBox (xmlLinkPtr lk)
{
  CardBoxType *cardBoxP = (CardBoxType*) xmlLinkGetData (lk);

  /*
   * Destroy/Delete the visited and unvisited list, including the data:
   * (perhaps could use lnode pool)
   */
  list_free_and_destroy (cardBoxP->expiredCards_listP);
  list_free_and_destroy (cardBoxP->dontNeedVisitCards_listP);

  /*
   * Don't forget to free the cardbox itself:
   */
  free (cardBoxP);
}


inline static int 
CompCardInBox (const void *key1, 
               const void *key2, 
               const void *data1,
               const void *data2,
               const void *)
{
  if ((uint32_t) key1 < (uint32_t) key2)
    return -1;

  if ((uint32_t) key1 > (uint32_t) key2)
    return 1;

  /*
   * tie in card id here, need to look at the file id:
   */
  if (likely (data2))
    {
      FlashCardInBoxType *cardInBox1P = (FlashCardInBoxType*)
        (FlashCardInBoxType*) lnode_get ((lnode_t*) data1);
      FlashCardInBoxType *cardInBox2P = (FlashCardInBoxType*)
        (FlashCardInBoxType*) lnode_get ((lnode_t*) data2);
      
      if (cardInBox1P->file_id < cardInBox2P->file_id)
          return -1;
      
      if (cardInBox1P->file_id > cardInBox2P->file_id)
          return 1;
    }
  return 0;
}



LeitnerModelClass::
LeitnerModelClass () :
  m_cardBoxCabinetFile_str (NULL),
  m_cardBoxCabinetCmpPath_str (NULL),
  m_cardBoxCabinetDirOnly_str (NULL),
  m_cabinetXmlState_enum (CABINET_XML_STATE_NONE),
  m_promoteCountThres (0),
  m_oriBox1Study_mins (0),
  m_oriHigherBoxesReview_mins (0),
  m_box1Study_mins (0),
  m_higherBoxesReview_mins (0),
  m_opened_flg (false),
  m_deckFileListP (NULL),
  m_flashCardObjA (NULL),
  m_tmpFlashCardFileIdx (0),
  m_cardBoxesListP (NULL),
  m_tmpCardBoxP (NULL),
  m_tmpCardBoxIdx (0),
  m_numCardBoxes (0),
  m_neverVisitedCard_int (0),
  m_cardBoxPA (NULL),
  m_numDrawCards_intA (NULL),
  m_cbCabinet_dictP (NULL),
  m_studyDynSeqNode_listP (NULL),
  m_curStudyCardNode_lnodeP (NULL),
  m_studySeqState_enum (STUDY_SEQ_STATE_NONE),
  m_studyCard_lnodePA (NULL),
  m_studyCardBox_idx (0),
  m_startStudyTime_mins (0),
  m_lastVisitedAt_mins (0),
  m_needSaving_flg (false),
  m_dontNeedVisit_cardsPA (NULL),
  m_expired_cardsPA (NULL),
  m_trashCardBox_idx (-1),
  m_numExpireds_intA (NULL),
  m_reviewNthBoxOnlg_idx (-1),
  m_lastCardBox_idx (-1),
  m_puaseTimeout_uint (DEFAULT_PAUSE_TIMEOUT_MIN),
  m_test_lnodePA (NULL),
  m_test_statusA (NULL),
  m_test_curIdx (-1),
  m_test_numCards (0),
  m_test_needSavingFlg (false),
  m_lastCardOriVisitedAt_mins (0),
  m_reachedEndOfCards_flg (false)
{
  fakeUseParam ();              // quiet the compiler
  fprintf (stderr, "calling constructor\n");
  m_cabinetXmlState_enum = CABINET_XML_STATE_NONE;
  m_deckFileListP = xmlListCreate (FreeDeckFile, NULL);
  m_flashCardObjA = NULL;

  m_opened_flg = false;
  m_tmpFlashCardFileIdx = 0;    // // for looping over callback function

  m_cardBoxesListP = xmlListCreate (FreeCardBox, NULL);
  m_tmpCardBoxP = NULL;         // for looping over callback function
  m_tmpCardBoxIdx = 0;          // // for looping over callback function
  m_numCardBoxes = 0;
  m_trashCardBox_idx = -1;
  m_cardBoxPA = NULL;
  m_numDrawCards_intA = NULL;

  m_cbCabinet_dictP = dict_create (DICTCOUNT_T_MAX, CompCardInBox, NULL);
  dict_allow_dupes (m_cbCabinet_dictP);

  m_studyDynSeqNode_listP = list_create (LISTCOUNT_T_MAX);

  m_studySeqState_enum = STUDY_SEQ_STATE_NONE;
  m_studyCard_lnodePA = NULL;
  m_studyCardBox_idx = 0;
  m_neverVisitedCard_int = 0;
  m_curStudyCardNode_lnodeP = NULL;

  m_cardBoxCabinetFile_str = NULL;
  m_cardBoxCabinetCmpPath_str = NULL;
  m_cardBoxCabinetDirOnly_str = NULL;

  m_startStudyTime_mins = 0;
  m_lastVisitedAt_mins = 0;
  m_needSaving_flg = false;

  m_dontNeedVisit_cardsPA = m_expired_cardsPA = NULL;
  srand (getpid ());            // for random seed

  m_promoteCountThres = 0;
  m_box1Study_mins = 0;
  m_higherBoxesReview_mins = 0;
  m_oriBox1Study_mins = 0;
  m_oriHigherBoxesReview_mins = 0;

  m_numExpireds_intA = NULL;
  m_reviewNthBoxOnlg_idx = -1;  // persistent even after a close()
  m_lastCardBox_idx = -1;

  m_puaseTimeout_uint = DEFAULT_PAUSE_TIMEOUT_MIN;

  m_test_lnodePA = NULL;
  m_test_statusA = NULL;
  m_test_curIdx = -1;
  m_test_numCards = 0;
  m_test_needSavingFlg = false;

  /* Restore last card to unprocessed: */
  m_lastCardOriVisitedAt_mins = 0;
  m_reachedEndOfCards_flg = false;
}


LeitnerModelClass::
~LeitnerModelClass ()
{
  fprintf (stderr, "calling destructor\n");
  Close (); 

  xmlListDelete (m_deckFileListP);
  m_deckFileListP = NULL;

  xmlListDelete (m_cardBoxesListP);
  m_cardBoxesListP = NULL;

  dict_destroy (m_cbCabinet_dictP);

  /*
   * Free and destroy cardbox pointer array:
   * (it was realloc in open, no freed in Close)
   */
  if (m_cardBoxPA)
    free (m_cardBoxPA);
  m_cardBoxPA = NULL;

  /*
   * Free the array that holds the number expired cards to draw per cardbox
   */
  delete [] m_numDrawCards_intA;

  /*
   * Free and destroy lnode pointer array:
   * (it was realloc in open, no freed in Close)
   */
  if (m_studyCard_lnodePA)
    free (m_studyCard_lnodePA);
  m_studyCard_lnodePA = NULL;

  /*
   * Free and destroy lnode pointer array:
   * (it was realloc in open, no freed in Close)
   */
  if (m_numExpireds_intA)
    free (m_numExpireds_intA);
  m_numExpireds_intA = NULL;


  /*
   * Destroy the study sequence list:
   * (do not free the data)
   */
  list_destroy (m_studyDynSeqNode_listP);
}


int LeitnerModelClass::
LoadFlashCards (const void *data, 
                const void *user)
{
  LeitnerModelClass *instancePtr = (LeitnerModelClass*) user;

  /* load the deck file: */
  char *fileName_str = (char*) data;
  fprintf (stderr, "**loading %s to %d\n", fileName_str,
           instancePtr->m_tmpFlashCardFileIdx);
  bool loadOp_flg = 
    instancePtr->m_flashCardObjA [instancePtr->m_tmpFlashCardFileIdx++].Load 
    (fileName_str, NULL);

  /* Check XML data integrity: */
  assert (loadOp_flg);          
  if (loadOp_flg == false)
    {
      fprintf (stderr, "fatal error in LoadFlashCards, can't continue\n");
      exit (-1);
    }

  // if (loadOp_flg)
  //   return 1;                   // keep walking
  // else
  //   return 0;                   // some problem
  return 1;                   // keep walking
}


bool LeitnerModelClass::
LoadDeckFiles ()
{
  fprintf (stderr, "calling LoadDeckFiles\n");
  /*
   * Allocate storage for array of deck files:
   */
  fprintf (stderr, "m_flashCardObjA is %p\n", m_flashCardObjA);
  assert (m_flashCardObjA == NULL);
  m_flashCardObjA = new FlashCardClass [xmlListSize (m_deckFileListP)];

  /*
   * load deck files:
   */
  m_tmpFlashCardFileIdx = 0;       // reset

  /* change current directory: */
  char *curDir_str = get_current_dir_name (); 
  assert (curDir_str);
  assert (m_cardBoxCabinetDirOnly_str);
  int retVal = chdir (m_cardBoxCabinetDirOnly_str);
  assert (retVal == 0);

  xmlListWalk (m_deckFileListP, LoadFlashCards, this);

  /* restore the current directory: */
  retVal = chdir (curDir_str);
  free (curDir_str);
  assert (retVal == 0);

  fprintf (stderr, "done calling LoadDeckFiles\n");
  return true;
}


void LeitnerModelClass::
UnloadDeckFiles ()
{
  int numFlashFiles_int = xmlListSize (m_deckFileListP);
  assert (numFlashFiles_int);
  assert (m_flashCardObjA);
  for (int idx = 0; idx < numFlashFiles_int; idx++)
    m_flashCardObjA [idx].Unload ();

  /*
   * Clean up for next load of deck files:
   */
  delete [] m_flashCardObjA;
  m_flashCardObjA = NULL;
}


void LeitnerModelClass::
InsertCardIntoBox (int                  in_file_id, 
                   FlashCardClass      *in_deckP,
                   const FlashCardType *in_flashCardP,
                   uint32_t             in_visitedAt_mins,
                   uint32_t             in_pCount,
                   bool                 in_visited_flg)
{
  assert (in_flashCardP);
  uint32_t card_id = in_flashCardP->wordUiq_id;

  /*
   * Check for already existing card in the cardbox cabinet:
   */
  FlashCardInBoxType dummyCardInBox;
  dummyCardInBox.file_id = in_file_id;
  lnode_t dummy_lnode;
  lnode_init (&dummy_lnode, &dummyCardInBox);
  dnode_t *dnodeP = dict_lookup (m_cbCabinet_dictP, (void*) card_id, 
                                 &dummy_lnode);
  if (dnodeP)
    {
      fprintf (stderr, "Duplicated card card_id=%d, file_id=%d\n",
               card_id, in_file_id);
      abort ();                 // TBF too harsh?
    }


  /*
   * Allocate storage for a FlashCardInBoxType:
   */
  FlashCardInBoxType *cardInBoxP = (FlashCardInBoxType*)
    malloc (sizeof (FlashCardInBoxType)); 
  cardInBoxP->flashCardP = in_flashCardP;
  cardInBoxP->file_id = in_file_id;
  cardInBoxP->visitedAt_mins = in_visitedAt_mins;
  cardInBoxP->promotionCount = in_pCount;
  cardInBoxP->oriPromotionCount = in_pCount;
  cardInBoxP->oriCardBoxIdx = m_numCardBoxes - 1;
  cardInBoxP->processed_enum = CARD_NOT_PROCESSED;
  cardInBoxP->cardBoxP = m_tmpCardBoxP;
  cardInBoxP->deckObjP = in_deckP;


  /*
   * Create a list node to hold the FlashCardInBoxType and insert it into
   * the corresponding list (visited or unviisted): (node data is
   * created, thus needs to be freed later in addition to node)
   */
  lnode_t* lnodeP = lnode_create ((void*) cardInBoxP); // need to free data
  assert (lnodeP);

  if (in_visited_flg)
    list_append (m_tmpCardBoxP->dontNeedVisitCards_listP, lnodeP);
  else
    list_append (m_tmpCardBoxP->expiredCards_listP, lnodeP);


  /*
   * Create a dictionary node to hold the FlashCardInBoxType and insert
   * into the dictionary: (node data is only linked, thus no needs to be
   * freed later, just free node)
   */
  dnodeP = dnode_create ((void*) lnodeP);
  dict_insert (m_cbCabinet_dictP, dnodeP, (void*) card_id); // key is card_id


  /*
   * Keep track of number of never visited cards:
   */
  if (in_visitedAt_mins == 0)
    m_neverVisitedCard_int ++;
}


void LeitnerModelClass::
InsertAllCardIntoBox (int      in_file_id, 
                      uint32_t in_visitedAt_mins,
                      uint32_t in_pCount,
                      bool     in_visited_flg)
{
  fprintf (stderr, "calling InsertAllCardIntoBox\n");
  int numCards_int;
  const FlashCardType **allCardsPA = 
    m_flashCardObjA [in_file_id].GetAllCards (&numCards_int);

  for (int idx = 0; idx < numCards_int; idx++)
    InsertCardIntoBox (in_file_id, m_flashCardObjA + in_file_id,
                       allCardsPA [idx], in_visitedAt_mins, 
                       in_pCount, in_visited_flg);
}



void LeitnerModelClass::
ProcessXmlNode (xmlTextReaderPtr in_readerP)
{
  int  nodeType_int = xmlTextReaderNodeType (in_readerP);

  /*
   * ignore #text :
   */
  if (nodeType_int == 14)
    return;


  /*
   * Get the name and value:
   */
  xmlChar *elmName = xmlTextReaderName (in_readerP);
  xmlChar *elmValue = xmlTextReaderValue (in_readerP);
  xmlChar *attName = NULL, *attValue = NULL;

  /*
   * State machine for CardBoxCabinet.dtd:
   */
  switch (m_cabinetXmlState_enum) 
    {
    case DECK_XML_STATE_NONE:
      /*
       * Look for "CardBoxCabinet" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "CardBoxCabinet") == 0))
        {
          m_cabinetXmlState_enum = CABINET_XML_STATE_CABINET;


          /* get the "promoteCountThres" and "lastUpdated" attributes: */
          uint32_t lastUpdated_mins = 0;
          while (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

              if (strcmp ((char*) attName, "promoteCountThres") == 0)
                m_promoteCountThres =  atoi ((char*) attValue);
              else if (strcmp ((char*) attName, "lastUpdated") == 0)
                lastUpdated_mins  = atoi ((char*) attValue); 
              else if (strcmp ((char*) attName, "box1Study") == 0)
                m_box1Study_mins  = atoi ((char*) attValue);
              else if (strcmp ((char*) attName, "higherBoxesReview") == 0)
                m_higherBoxesReview_mins = atoi ((char*) attValue); 
              else 
                {
                  fprintf (stderr, "xml format problem?? '%s'\n",
                           (char*) attName);
                  abort ();    // maybe too harsh?  TBF
                }
            }
          
          /* check for optional study duration limites: */
          if (m_box1Study_mins == 0)
            m_box1Study_mins = 35;
          if (m_higherBoxesReview_mins == 0)
            m_higherBoxesReview_mins = 45;

          /* save the original time limits: */
          m_oriBox1Study_mins = m_box1Study_mins;
          m_oriHigherBoxesReview_mins = m_higherBoxesReview_mins;


          /* clean ups: */
          xmlFree (attName);
          xmlFree (attValue);
        }
      break;
    case CABINET_XML_STATE_CABINET:  // no break, followed by deckfile element
    case CABINET_XML_STATE_DECKFILE: // one or more
      /*
       * Look for "CardDeckFile" start element:
       */
//       fprintf (stderr, "elmName = '%s'\n", (char*) elmName);
      if (strcmp ((char*) elmName, "#comment") == 0) // skip comments
        break;

      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "CardDeckFile") == 0))
        {
          m_cabinetXmlState_enum = CABINET_XML_STATE_DECKFILE;
          
          /* get the "file" attribute: */
          xmlTextReaderMoveToNextAttribute (in_readerP);
          attName = xmlTextReaderName (in_readerP);
          attValue = xmlTextReaderValue (in_readerP);
          
          assert (strcmp ((char*) attName, "file") == 0);
//           fprintf (stderr, "file = '%s'\n", attValue);

          /* allocate storage and insert into the list: */
          xmlListPushBack (m_deckFileListP, attValue); // don't xmlFree attValue
          
          /* clean ups: but don't xmlFree (attValue) */
          xmlFree (attName);

          break;                // break inside.
        }
      // no break here
      assert (xmlListSize (m_deckFileListP));
      
      /*
       * Done with all CardDeckFiles:
       * (a good place to allocate and read in the flash cards)
       */
      m_cabinetXmlState_enum = CABINET_XML_STATE_CARDBOX;
      LoadDeckFiles ();
      
      
    case CABINET_XML_STATE_CARDBOX: // one or more
      /*
       * Look for "CardBox" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          (strcmp ((char*) elmName, "CardBox") == 0))
        {
          m_cabinetXmlState_enum = CABINET_XML_STATE_CARDBOX;


          /* allocate storage for a new cardbox and
             insert it into the list: */
          m_tmpCardBoxP = (CardBoxType*) malloc (sizeof (CardBoxType));
          m_tmpCardBoxP->expiredCards_listP = list_create (LISTCOUNT_T_MAX);
          m_tmpCardBoxP->dontNeedVisitCards_listP = 
            list_create (LISTCOUNT_T_MAX);
          xmlListPushBack (m_cardBoxesListP, m_tmpCardBoxP);
          m_numCardBoxes++;

          /* get the "expPeriodHrs", "nextExpDate", "expired_flg" attributes: */
          while (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

//               fprintf (stderr, "attribute '%s' = '%s'\n", attName, attValue);
              if (strcmp ((char*) attName, "expPeriodHrs") == 0)
                {
                  m_tmpCardBoxP->expirationPeriod_hrs =  (uint32_t)
                    atoi ((char*) attValue); // trash cardbox has -1
                  if ((uint32_t) -1 == m_tmpCardBoxP->expirationPeriod_hrs)
                    {
                      assert (m_trashCardBox_idx == -1); // only one trash
                      m_trashCardBox_idx = m_numCardBoxes - 1; // trash
                      fprintf (stderr, "trash carbox is at %d\n",
                               m_trashCardBox_idx);
                      assert (m_trashCardBox_idx > 0); // not cardbox1
                    }
                }
              else 
                {
                  fprintf (stderr, "xml format problem?? '%s'\n",
                           (char*) attName);
                  abort ();    // maybe too harsh?  TBF
                }

              /* clean ups: */
              xmlFree (attName);
              xmlFree (attValue);
            }
        }
      /*
       * Look for "VisitedCards" or "UnVisitedCards" start element:
       */
      if ((xmlTextReaderNodeType (in_readerP) == 1) &&
          ((strcmp ((char*) elmName, "VisitedCards") == 0) ||
           (strcmp ((char*) elmName, "UnVisitedCards") == 0)))
        {
          bool visited_flg;
          if (strcmp ((char*) elmName, "UnVisitedCards") == 0)
            visited_flg = false;
          else
            visited_flg = true;

          /* get the "file_id", "card_id", "visitedAt", "pCount" attributes: */
          uint32_t file_id = 0, card_id = 0, visitedAt_mins = 0, pCount = 0;
          bool allIds_flg = false;
          while (xmlTextReaderMoveToNextAttribute (in_readerP))
            {
              attName = xmlTextReaderName (in_readerP);
              attValue = xmlTextReaderValue (in_readerP);

//               fprintf (stderr, "attribute '%s' = '%s'\n", attName, attValue);
              if (strcmp ((char*) attName, "file_id") == 0)
                file_id =  atoi ((char*) attValue);
              else if (strcmp ((char*) attName, "card_id") == 0)
                {
                  if (strcmp ((char*) attValue, "_all") == 0)
                    allIds_flg = true;
                  else
                    card_id = atoi ((char*) attValue + 1); // skip '_'
                }
              else if (strcmp ((char*) attName, "visitedAt") == 0)
                visitedAt_mins  = atoi ((char*) attValue); 
              else if (strcmp ((char*) attName, "pCount") == 0)
                pCount  = atoi ((char*) attValue); 
              else 
                {
                  fprintf (stderr, "xml format problem?? '%s'\n",
                           (char*) attName);
                  abort ();    // maybe too harsh?  TBF
                }

              /* clean ups: */
              xmlFree (attName);
              xmlFree (attValue);
            }

          /*
           * insert cards into the UnVisitedCards or VisitedCards list:
           */
          if ((int) file_id >= xmlListSize (m_deckFileListP))
            {
              fprintf (stderr, "not enough CardDeckFile elements\n");
              abort ();         // TBF, too harsh?
            }
          if (allIds_flg)
            InsertAllCardIntoBox (file_id, visitedAt_mins, pCount, visited_flg);
          else
            {
              const FlashCardType* flashCardP = 
                m_flashCardObjA [file_id].GetCard (card_id);
              assert (flashCardP);
              InsertCardIntoBox (file_id, m_flashCardObjA + file_id,
                                 flashCardP, visitedAt_mins, 
                                 pCount, visited_flg);
            }
        }

      break;
    default:
      break;
    }


  /*
   * clean ups:
   */
  xmlFree (elmName);
  xmlFree (elmValue);
}




int LeitnerModelClass::
InsertCardBoxInArray (const void *data, 
                      const void *user)
{
  LeitnerModelClass *instancePtr = (LeitnerModelClass*) user;

  instancePtr->m_cardBoxPA [instancePtr->m_tmpCardBoxIdx++] =
    (CardBoxType*) data;
  
  return 1;                     // keep walking
}




bool LeitnerModelClass::
Open (const char *in_cardBoxCabinetFile_str)
{
  if (m_opened_flg)
    return false;


  /*
   * 0.) Obtain the path to the cardboxCabine file:
   */
  assert (m_cardBoxCabinetCmpPath_str == NULL);
  m_cardBoxCabinetCmpPath_str = strdup (in_cardBoxCabinetFile_str);
  m_cardBoxCabinetDirOnly_str = strdup (dirname (m_cardBoxCabinetCmpPath_str)); 
  assert (m_cardBoxCabinetDirOnly_str);
  fprintf (stderr, "m_cardBoxCabinetDirOnly_str is %s\n",
           m_cardBoxCabinetDirOnly_str);
  
  /*
   * 1.) Open the CardBoxCabinet file for XML parsing:
   */
  m_cardBoxCabinetFile_str = strdup (in_cardBoxCabinetFile_str);
  xmlTextReaderPtr readerP = 
    xmlNewTextReaderFilename (in_cardBoxCabinetFile_str);
  if (readerP == NULL)
    {
      fprintf (stderr, "could not read cabinet file %s\n", 
               in_cardBoxCabinetFile_str);
      return false;
    }


  /*
   * Parse the cabinet file and fill data:
   */
  m_cabinetXmlState_enum = CABINET_XML_STATE_NONE;
  int ret = xmlTextReaderRead (readerP);
  while (ret == 1)
    {
      ProcessXmlNode (readerP);
      ret = xmlTextReaderRead (readerP);
    }
  /* clean up: */
  xmlTextReaderClose (readerP);
  xmlFreeTextReader (readerP);
  

  /*
   * Create a Trash cardbox if not in the file CardBoxCabinet file:
   */
  if (m_trashCardBox_idx == -1)
    {
      fprintf (stderr, "Creating the trash cardbox\n");
      /* allocate storage for a new cardbox and insert it into the list: */
      m_tmpCardBoxP = (CardBoxType*) malloc (sizeof (CardBoxType));
      m_tmpCardBoxP->expiredCards_listP = list_create (LISTCOUNT_T_MAX);
      m_tmpCardBoxP->dontNeedVisitCards_listP = 
        list_create (LISTCOUNT_T_MAX);
      xmlListPushBack (m_cardBoxesListP, m_tmpCardBoxP);
      m_numCardBoxes++;

      /* mark as trash cardbox: */
      m_tmpCardBoxP->expirationPeriod_hrs = (uint32_t) -1;
      m_trashCardBox_idx = m_numCardBoxes - 1; // trash
    }


  /*
   * Re-Create a cardbox pointer array for easy access:
   */
  assert (m_numCardBoxes == xmlListSize (m_cardBoxesListP));
  m_cardBoxPA = (CardBoxType**) realloc (m_cardBoxPA, // be careful
                                         sizeof (CardBoxType*) *
                                         m_numCardBoxes);
  assert (m_cardBoxPA);
  m_numExpireds_intA = (int*) realloc (m_numExpireds_intA, // be careful
                                       sizeof (int) *
                                       m_numCardBoxes);
  assert (m_numExpireds_intA);
  m_tmpCardBoxIdx = 0;          // reset
  xmlListWalk (m_cardBoxesListP, InsertCardBoxInArray, this);
  

  /*
   * Get the last cardbox that is not trash box:
   */
  if (m_trashCardBox_idx != -1)
    {
      for (int idx = m_numCardBoxes - 1; idx >= 0; idx --)
        {
          if (idx != m_trashCardBox_idx)
            {
              m_lastCardBox_idx = idx;
              break;
            }
        }
    }
  else
    m_lastCardBox_idx = m_numCardBoxes - 1;

  assert ((m_lastCardBox_idx >= 0) && (m_lastCardBox_idx < m_numCardBoxes));


  /*
   * Compute the number of expired cards to draw from each cardbox:
   */
  delete [] m_numDrawCards_intA;
  m_numDrawCards_intA = new int [m_numCardBoxes];
  for (int idx = 0; idx < m_numCardBoxes; idx ++)
    {
      m_numDrawCards_intA [idx] = -(idx - m_numCardBoxes);
      fprintf (stderr, "cardbox %d, draw %d expired cards\n",
               idx, m_numDrawCards_intA [idx]);
    }


  /*
   * Start the study session:
   */
  m_startStudyTime_mins = GetMinSinceEpoch ();
  m_lastVisitedAt_mins = m_startStudyTime_mins;
  fprintf (stderr, "study started at minutes is %d\n", m_startStudyTime_mins);


  /*
   * Check and Move all expired cards:
   */
  MoveAllExpiredCards (m_startStudyTime_mins);

  
  /*
   * Recreate a lnode pointer array for study sequence:
   */
  m_studyCard_lnodePA = (lnode_t**) realloc (m_studyCard_lnodePA, // be careful
                                             sizeof (lnode_t*) *
                                             m_numCardBoxes);
  assert (m_studyCard_lnodePA);
  for (int idx = 0; idx < m_numCardBoxes; idx++)
    m_studyCard_lnodePA [idx] = NULL;


  /*
   * Clean up and set application state:
   */
  fprintf (stderr, "there are %d deck files\n", xmlListSize (m_deckFileListP));
  m_opened_flg = true;
  return true;
}


static int 
WirteFlashCardFileNames (const void *data, 
                         const void *user)
{
  FILE *fileP = (FILE*) user;
  const char *fileName_str = (char*) data;

  fprintf (fileP, "  <CardDeckFile file=\"%s\"/>\n", fileName_str);
  return 1;                     // keep walking
}



static void
CreateKnuthShuffle (int *indexA,
                    int  numElms)
{
  /* go through positions from 1 to n - 1: */
  for (int idx = 0; idx < (numElms - 1); idx ++)
    {
      /* swap with position from i to n: */
      int s_idx = idx + (rand() % (numElms - idx));

      int tmp = indexA [idx];
      indexA [idx] = indexA [s_idx];
      indexA [s_idx] = tmp;
    }
}


static void
SaveCards (FILE                *fileP,
           list_t              *cards_listP,
           const char          *cardType_str,
           FlashCardInBoxType **in_cardsPA)
{
  int numCards = (int) list_count (cards_listP);
  FlashCardInBoxType *cardInBoxP;

  /*
   * Are cards saved in the order of list?
   */
  if (in_cardsPA == NULL)
    {
      lnode_t *lnodeP = list_first (cards_listP);
      for (int v_idx = 0; v_idx < numCards; v_idx++)
        {
          cardInBoxP = (FlashCardInBoxType*) lnode_get (lnodeP);
            
          if (cardInBoxP->promotionCount)
            fprintf (fileP, "    <%s file_id=\"%d\""
                     " card_id=\"_%d\" visitedAt=\"%d\" pCount=\"%d\"/>\n",
                     cardType_str,
                     cardInBoxP->file_id, 
                     cardInBoxP->flashCardP->wordUiq_id,
                     cardInBoxP->visitedAt_mins,
                     cardInBoxP->promotionCount);
          else
            fprintf (fileP, "    <%s file_id=\"%d\""
                     " card_id=\"_%d\" visitedAt=\"%d\"/>\n",
                     cardType_str,
                     cardInBoxP->file_id, 
                     cardInBoxP->flashCardP->wordUiq_id,
                     cardInBoxP->visitedAt_mins);
          
          lnodeP = list_next (cards_listP, lnodeP);
        }
      return;
    }


  /*
   * create permutation of indexes:
   */
  int permuteIdxA [numCards];
  for (int idx = 0; idx < numCards; idx++)
    permuteIdxA [idx] = idx;
  CreateKnuthShuffle (permuteIdxA, numCards);

  /*
   * Save in the random order:
   */
  for (int v_idx = 0; v_idx < numCards; v_idx++)
    {
      cardInBoxP = in_cardsPA [permuteIdxA [v_idx]];
      if (cardInBoxP->promotionCount)
        fprintf (fileP, "    <%s file_id=\"%d\""
                 " card_id=\"_%d\" visitedAt=\"%d\" pCount=\"%d\"/>\n",
                 cardType_str,
                 cardInBoxP->file_id, 
                 cardInBoxP->flashCardP->wordUiq_id,
                 cardInBoxP->visitedAt_mins,
                 cardInBoxP->promotionCount);
      else
        fprintf (fileP, "    <%s file_id=\"%d\""
                 " card_id=\"_%d\" visitedAt=\"%d\"/>\n",
                 cardType_str,
                 cardInBoxP->file_id, 
                 cardInBoxP->flashCardP->wordUiq_id,
                 cardInBoxP->visitedAt_mins);
    }
}


static void
SaveRandomizeUnvisitedCards (FILE                *fileP,
                             int                  in_numCards,
                             FlashCardInBoxType **in_cardsPA)
{
  FlashCardInBoxType *cardInBoxP;

  /*
   * Save as is for expired cards (VistedAt  != "0"):
   */
  int zeroVisited_idx;
  for (zeroVisited_idx = 0; zeroVisited_idx < in_numCards; zeroVisited_idx++)
    {
      cardInBoxP = in_cardsPA [zeroVisited_idx];
      if (likely (cardInBoxP->visitedAt_mins != 0)) // front portions of list
        {
          if (cardInBoxP->promotionCount)
            fprintf (fileP, "    <UnVisitedCards file_id=\"%d\""
                     " card_id=\"_%d\" visitedAt=\"%d\" pCount=\"%d\"/>\n",
                     cardInBoxP->file_id, 
                     cardInBoxP->flashCardP->wordUiq_id,
                     cardInBoxP->visitedAt_mins,
                     cardInBoxP->promotionCount);
          else
            fprintf (fileP, "    <UnVisitedCards file_id=\"%d\""
                     " card_id=\"_%d\" visitedAt=\"%d\"/>\n",
                     cardInBoxP->file_id, 
                     cardInBoxP->flashCardP->wordUiq_id,
                     cardInBoxP->visitedAt_mins);
        }
      else
        break;
    }


  /*
   * Are there any zero visted cards?
   */
  if (zeroVisited_idx == in_numCards)
    return;
  
  /*
   * create permutation of indexes for the zero visited cards:
   */
  int numZeroCards = in_numCards - zeroVisited_idx;
  int permuteIdxA [numZeroCards];
  for (int idx = 0; idx < numZeroCards; idx++)
    permuteIdxA [idx] = idx + zeroVisited_idx;
  CreateKnuthShuffle (permuteIdxA, numZeroCards);

  /*
   * Save the zero visited cards in random order:
   */
  for (int v_idx = 0; v_idx < numZeroCards; v_idx++)
    {
      cardInBoxP = in_cardsPA [permuteIdxA [v_idx]];
      fprintf (fileP, "    <UnVisitedCards file_id=\"%d\""
               " card_id=\"_%d\" visitedAt=\"0\"/>\n",
               cardInBoxP->file_id, 
               cardInBoxP->flashCardP->wordUiq_id);
      assert (cardInBoxP->visitedAt_mins == 0);
    }
}


void LeitnerModelClass::
SaveTestSession (const char *in_testFile_str,
                 bool        in_isClosing_flg)
{
  FILE *fileP = fopen (in_testFile_str, "wb");
  if (fileP == NULL)
    return;


  // /*
  //  * Reset the last visited card to "not_visited" unless that card is
  //  * also the last card the entire test.  This is because by default,
  //  * when the card is first presented, it is set to "nope".
  //  */
  // if (m_test_curIdx != (m_test_numCards - 1))
  //   m_test_statusA [m_test_curIdx] = TEST_NOT_VISITED;

  /*
   * Reset the last visited card to "not_visited" if saved during closing
   * the application.  
   */
  if (in_isClosing_flg)
    m_test_statusA [m_test_curIdx] = TEST_NOT_VISITED;

  char     status_str [20];
  int      fileId_int;
  uint32_t cardId_u32;
  FlashCardInBoxType* cardInBoxP;
  for (int idx = 0; idx < m_test_numCards; idx++)
    {
      /* prepare the status string: */
      switch (m_test_statusA [idx])
        {
        case TEST_NOT_VISITED:
          sprintf (status_str, "not_visited");
          break;
        case TEST_NOPE:
          sprintf (status_str, "nope");
          break;
        case TEST_YEP:
          sprintf (status_str, "yep");
          break;
        }

      /* get the file and card ids: */
      assert (m_test_lnodePA [idx]);
      cardInBoxP = (FlashCardInBoxType*) lnode_get (m_test_lnodePA [idx]);
      fileId_int = cardInBoxP->file_id;
      cardId_u32 = cardInBoxP->flashCardP->wordUiq_id;

      /* write entry to the file: */
      fprintf (fileP, "file_id=\"%d\" card_id=\"_%d\" status=\"%s\"\n",
               fileId_int, cardId_u32, status_str);

    }
  fclose (fileP);
}

void LeitnerModelClass::
RestoreLastCardToUnprocessed ()
{
  fprintf (stderr, "calling DontSaveLastCard\n");


  /* Get the last visited card: */
  lnode_t *curStudyCard_lnodeP = 
    list_first ((m_cardBoxPA [0])->dontNeedVisitCards_listP);
  FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
    lnode_get (curStudyCard_lnodeP);
  // PrintCardInBoxInfo (cardInBoxP);

              
  /* Has the user promoted the last card? */
  if (cardInBoxP->cardBoxP != m_cardBoxPA [0]) // yes, already promoted
    return;

  /* The last card was demoted by default, the user may not want to save
     the card as marked demoted.  So, we restore the card to original
     cardbox. */
  /* Mark as not-processed: */
  cardInBoxP->processed_enum = CARD_NOT_PROCESSED; 
  cardInBoxP->visitedAt_mins = m_lastCardOriVisitedAt_mins;
  cardInBoxP->promotionCount = cardInBoxP->oriPromotionCount;
  int fromBox_idx = cardInBoxP->oriCardBoxIdx;
  if (fromBox_idx == 0)
    return;                     // the card was originally in cardbox #0

  /* 
   * "promote" back to its original cardbox: 
   */
  /* update the study sequence before deletion from the list: */
  list_t *card_listP = (m_cardBoxPA[0])->dontNeedVisitCards_listP;
  // fprintf (stderr, "?????? %p is in %p\n",
  //          curStudyCard_lnodeP, (m_cardBoxPA [0])->dontNeedVisitCards_listP);
  assert (list_contains (card_listP, curStudyCard_lnodeP));
  m_studyCard_lnodePA [0] = list_prev (card_listP, curStudyCard_lnodeP);

  /* remove the card from its current list: */
  list_delete (m_cardBoxPA[0]->dontNeedVisitCards_listP,
               curStudyCard_lnodeP);

  /* update the card's box pointer to higher CardBox1: */
  cardInBoxP->cardBoxP = m_cardBoxPA [fromBox_idx];

  /* move (append) to the original cardbox's expiredCards_listP: */
  list_prepend (m_cardBoxPA [fromBox_idx]->expiredCards_listP,
                curStudyCard_lnodeP);

  // fprintf (stderr, "after\n");
  // PrintCardInBoxInfo (cardInBoxP);


  /*
   * Update the m_studyDynSeqNode_listP list: (not necessary...)
   */
  lnode_t *curStudyCardNode_lnodeP = list_last (m_studyDynSeqNode_listP);
  assert (curStudyCardNode_lnodeP);
  
  curStudyCard_lnodeP = (lnode_t*)
    lnode_get ((lnode_t*) lnode_get (curStudyCardNode_lnodeP));

  /* remove the card from the m_studyDynSeqNode_listP list: */
  assert (list_contains (m_studyDynSeqNode_listP, curStudyCardNode_lnodeP));
  list_delete (m_studyDynSeqNode_listP, curStudyCardNode_lnodeP);
  lnode_destroy (curStudyCardNode_lnodeP);
}


bool LeitnerModelClass::
Save (const char* in_cabinetFile_str,
      bool        in_exportDkf_flg,
      bool        in_isClosing_flg) // if set to true, must call Close() next
{
  if (m_opened_flg == false)
    {
      fprintf (stderr, "nothing to save\n");
      return true;
    }

  assert (m_cardBoxCabinetFile_str);
  FILE *fileP;
  if (in_cabinetFile_str == NULL)
    in_cabinetFile_str = m_cardBoxCabinetFile_str;
  fileP = fopen (in_cabinetFile_str, "wb");

  if (fileP == NULL)
    {
      fprintf (stderr, "could not open file for saving: %s\n", 
               m_cardBoxCabinetFile_str);
      return false;
    }

  /*
   * Don't save the last card, if user has not process it:
   */
  if (in_isClosing_flg && (m_reachedEndOfCards_flg == false))
    RestoreLastCardToUnprocessed ();


  /*
   * Export to a single Dkf file.
   */
  if (in_exportDkf_flg)
    {
      char *exportFile_str = (char*) malloc (strlen (in_cabinetFile_str) +
                                             sizeof ("_exp") + 1); // add '\0'
      strcpy (exportFile_str, in_cabinetFile_str);
      char *fileExt_chrP = strstr (exportFile_str, ".cbx");
      assert (fileExt_chrP);
      if (fileExt_chrP == NULL)
        fprintf (stderr, "Could not open default test file for saving\n");
      else
        {
          sprintf (fileExt_chrP, "_exp.dkf"); // overwrite file extension
          ExportToDeck (exportFile_str);  
        }
    }



  /*
   * Write the file header block:
   */
  fprintf (fileP, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  fprintf (fileP, "\n");
  fprintf (fileP, "<!-- Written by leitnerSysModel -->\n");
  fprintf (fileP, "\n");

  /*
   * Write the CardBoxbCabinet header:
   */
  uint32_t lastUpdated_mins = GetMinSinceEpoch ();

  fprintf (fileP, "<!DOCTYPE CardBoxCabinet SYSTEM \"cardBoxCabinet.dtd\">\n");
  fprintf (fileP, 
           "<CardBoxCabinet promoteCountThres=\"%d\" lastUpdated=\"%d\"\n"
           "                box1Study=\"%d\" higherBoxesReview=\"%d\">\n",
           m_promoteCountThres, lastUpdated_mins,
           m_oriBox1Study_mins, m_oriHigherBoxesReview_mins);

  /*
   * Write the file block:
   */
  xmlListWalk (m_deckFileListP, WirteFlashCardFileNames, fileP);

  /*
   * Write the CardBox blocks:
   */
  list_t *cards_listP ;
  for (int idx = 0; idx < m_numCardBoxes; idx++)
    {
      CardBoxType *cardBoxP = m_cardBoxPA [idx];

      /* 
       * Start the CardBox: 
       */
      fprintf (fileP, "  <CardBox expPeriodHrs=\"%d\">\n",
               cardBoxP->expirationPeriod_hrs);

      FlashCardInBoxType **dontNeedVisit_cardsPA = NULL; // use builtin memory
      FlashCardInBoxType **expired_cardsPA = NULL; // use builtin memory
      int numExpired_int;
      GetAllCardsInBox (idx, NULL, &dontNeedVisit_cardsPA,
                        &numExpired_int, &expired_cardsPA);
      // GetAllCardsInBox (idx, NULL, &dontNeedVisit_cardsPA,
      //                   NULL, NULL);

      /* Write the visited (dontNeedVisitCards_listP) cards: */
      cards_listP = cardBoxP->dontNeedVisitCards_listP;
      SaveCards (fileP, cards_listP, "VisitedCards", dontNeedVisit_cardsPA);

      /* 
       * Write the unvisited (expiredCards_listP) cards: 
       */
      SaveRandomizeUnvisitedCards (fileP, numExpired_int, expired_cardsPA);

      /*
       * close the CardBox: 
       */
      fprintf (fileP, "  </CardBox>\n");
    }

  /*
   * close the CardBoxbCabinet:
   */
  fprintf (fileP, "</CardBoxCabinet>\n");
  fclose (fileP);

  /*
   * Save the test session:
   */
  if (m_test_numCards)
    {
      char *testFile_str = strdup (in_cabinetFile_str);
      char *fileExt_chrP = strstr (testFile_str, ".cbx");
      assert (fileExt_chrP);
      if (fileExt_chrP == NULL)
        fprintf (stderr, "Could not open default test file for saving\n");
      else
        {
          sprintf (fileExt_chrP, ".tst"); // overwrite file extension
          SaveTestSession (testFile_str, in_isClosing_flg);
        }
      free (testFile_str);
    }

  /*
   * update the need saving state:
   */
  m_needSaving_flg = false;
  m_test_needSavingFlg = false;

  return true;
}


bool LeitnerModelClass::
ExportToDeck (const char *in_deckFile_str)
{
  assert (m_flashCardObjA);
  if (m_flashCardObjA == NULL)
    {
      fprintf (stderr, "nothing to export\n");
      return false;
    }

  DeckSavingClass deckSaveObj (in_deckFile_str,
                               m_flashCardObjA[0].GetSoundPath ());

  for (int box_idx = 0; box_idx < m_numCardBoxes; box_idx++)
    {
      if (unlikely (box_idx == m_trashCardBox_idx)) // skip trash card box
        continue;

      list_t *cards_listP = m_cardBoxPA [box_idx]->dontNeedVisitCards_listP;
      lnode_t *lnodeP = list_first (cards_listP);
      while (lnodeP)
        {
          const FlashCardType *flashCardP = ((FlashCardInBoxType*) 
                                             lnode_get (lnodeP))->flashCardP;
          deckSaveObj.InsertCard (flashCardP);
          lnodeP = list_next (cards_listP, lnodeP);
        }
      
      cards_listP = m_cardBoxPA [box_idx]->expiredCards_listP;
      lnodeP = list_first (cards_listP);
      while (lnodeP)
        {
          const FlashCardType *flashCardP = ((FlashCardInBoxType*) 
                                             lnode_get (lnodeP))->flashCardP;
          deckSaveObj.InsertCard (flashCardP);
          lnodeP = list_next (cards_listP, lnodeP);
        }
    }


  return true;
}


bool LeitnerModelClass::
NeedSave ()
{
  return m_needSaving_flg || m_test_needSavingFlg;
}


void LeitnerModelClass::
Close ()
{
  if (m_opened_flg)
    {
      /*
       * Free and unload deck files:
       * (must be first, since it depends on sizeof m_deckFileListP)
       */
      UnloadDeckFiles (); 

      /*
       * Free but not destroy deck file name list:
       */
      xmlListClear (m_deckFileListP);
      assert (xmlListSize (m_deckFileListP) == 0);

      /*
       * Free but not destroy cardbox list:
       */
      xmlListClear (m_cardBoxesListP);
      assert (xmlListSize (m_cardBoxesListP) == 0);

      /*
       * Free but not destroy cardbox cabinet dictionary:
       */
      dict_free_nodes (m_cbCabinet_dictP); // use dnode pool?
      assert (dict_count (m_cbCabinet_dictP) == 0);

      /*
       * Free the lnodes of the study sequence list:
       * (do not free the data)
       */
      list_destroy_nodes (m_studyDynSeqNode_listP);
      assert (list_count (m_studyDynSeqNode_listP) == 0);

      /*
       * free the file names:
       */
      free (m_cardBoxCabinetFile_str);
      m_cardBoxCabinetFile_str = NULL;

      free (m_cardBoxCabinetCmpPath_str);
      m_cardBoxCabinetCmpPath_str = NULL;
      free (m_cardBoxCabinetDirOnly_str);
      m_cardBoxCabinetDirOnly_str = NULL; 


      /*
       * Destroy test cards and status arrays:
       * (do not free the shared data)
       */
      if (m_test_lnodePA)
        free (m_test_lnodePA);
      m_test_lnodePA = NULL;
      if (m_test_statusA) 
        free (m_test_statusA);
      m_test_statusA = NULL;

      /*
       * Reset the test state: 
       */
      m_test_curIdx = -1;
      m_test_numCards = 0;
      m_test_needSavingFlg = false;

    }

  /*
   * Reset state machines:
   */
  m_cabinetXmlState_enum = CABINET_XML_STATE_NONE;
  m_studySeqState_enum = STUDY_SEQ_STATE_NONE;
  m_studyCardBox_idx = 0;
  m_numCardBoxes = 0;
  m_trashCardBox_idx = -1;
  m_neverVisitedCard_int = 0;
  m_curStudyCardNode_lnodeP = NULL;

  m_tmpFlashCardFileIdx = 0;    // // for looping over callback function
  m_tmpCardBoxP = NULL;         // for looping over callback function
  m_tmpCardBoxIdx = 0;          // // for looping over callback function

  m_opened_flg = false;
  m_startStudyTime_mins = 0;
  m_lastVisitedAt_mins = 0;
  m_needSaving_flg = false;

  m_promoteCountThres = 0;
  m_box1Study_mins = 0;
  m_higherBoxesReview_mins = 0;
  m_oriBox1Study_mins = 0;
  m_oriHigherBoxesReview_mins = 0;

  m_lastCardOriVisitedAt_mins = 0;
  m_reachedEndOfCards_flg = false;

  /*
   * Free build-in memory:
   */
  Free (m_dontNeedVisit_cardsPA);
  Free (m_expired_cardsPA);
}


void LeitnerModelClass::
MoveToExpiredCards (CardBoxType *in_cardBoxP,
                    MoveModeEnum in_moveMode_enum,
                    uint32_t     in_curTime_mins)
{
  list_t *expd_listP = in_cardBoxP->expiredCards_listP;
  list_t *notExpd_listP = in_cardBoxP->dontNeedVisitCards_listP;
  int numElm_int = list_count (notExpd_listP);
  lnode_t *noExpd_lnodeP = list_first (notExpd_listP);
  FlashCardInBoxType *cardInBoxP = NULL;

  switch (in_moveMode_enum)
    {
      /*
       * Checks and moves the notExpd_listP elements to the beginning of
       * expd_listP:
       */
    case MOVE_PREPEND:
      for (int idx = 0; idx < numElm_int; idx++)
        {
          assert (noExpd_lnodeP);
          cardInBoxP = (FlashCardInBoxType*) lnode_get (noExpd_lnodeP);
          if ((in_curTime_mins - cardInBoxP->visitedAt_mins) >= // could be 0
              (in_cardBoxP->expirationPeriod_hrs * 60)) // card expired
            {
              lnode_t *lnodeP = list_next (notExpd_listP, noExpd_lnodeP);
              list_delete (notExpd_listP, noExpd_lnodeP); // this will reverse
              list_prepend (expd_listP, noExpd_lnodeP);   // the order, oh well
              noExpd_lnodeP = lnodeP;
            }
          else
            noExpd_lnodeP = list_next (notExpd_listP,  // TC 2009-08-23 --
                                       noExpd_lnodeP); // was bug! 
        }
      break;

      /*
       * Checks and moves the notExpd_listP elements to the end of
       * expd_listP:
       */
    case MOVE_APPEND:
      for (int idx = 0; idx < numElm_int; idx++)
        {
          assert (noExpd_lnodeP);
          cardInBoxP = (FlashCardInBoxType*) lnode_get (noExpd_lnodeP);
          if ((in_curTime_mins - cardInBoxP->visitedAt_mins) >
              (in_cardBoxP->expirationPeriod_hrs * 60)) // card expired
            {
              // fprintf (stderr, "EXPIRED!\n");
              lnode_t *lnodeP = list_next (notExpd_listP, noExpd_lnodeP);
              list_delete (notExpd_listP, noExpd_lnodeP);
              list_append (expd_listP, noExpd_lnodeP); 
              noExpd_lnodeP = lnodeP;
            }
          else
            noExpd_lnodeP = list_next (notExpd_listP,  // TC 2009-08-23 --
                                       noExpd_lnodeP); // was bug! 
        }
      break;
    default:
      break;
    }
}


void LeitnerModelClass::
MoveAllExpiredCards (uint32_t in_curTime_mins)
{
  /*
   * Prepend card that just expires into the expiredCards_listP for
   * CardBox1:
   */
  MoveToExpiredCards (m_cardBoxPA [0], MOVE_PREPEND, in_curTime_mins);

  /*
   * Append card that just expires into the expiredCards_listP for
   * higher CardBoxes:
   */
  fprintf (stderr, "m_numCardBoxes is %d\n", m_numCardBoxes);
  for (int idx = 1; idx < m_numCardBoxes; idx++)
    MoveToExpiredCards (m_cardBoxPA [idx], MOVE_APPEND, in_curTime_mins);
}



FlashCardInBoxType* LeitnerModelClass::
GetNextCard ()
{
  /*
   * Check to see if the current card has been processed:
   */
  FlashCardInBoxType *curCardInBoxP = GetCurCard ();
  if (curCardInBoxP && (curCardInBoxP->processed_enum == CARD_NOT_PROCESSED))
    {
      fprintf (stderr, "You must process the current card first\n");
      return curCardInBoxP;
    }


  /*
   * Resume timer and get current time for visitedAt:
   */
  ResumeTimer ();  
  uint32_t curTimeAt_mins = GetMinSinceEpoch ();

  bool doneStudy_flg = false;
  lnode_t *studyCard_lnodeP = NULL;


  /*
   * Retrieve the next card from already traversed list:
   */
  if (m_curStudyCardNode_lnodeP &&
      (list_last (m_studyDynSeqNode_listP) != m_curStudyCardNode_lnodeP))
    {
      fprintf (stderr, "Retrieving from already traversed list\n");

      m_curStudyCardNode_lnodeP = 
        list_next (m_studyDynSeqNode_listP, m_curStudyCardNode_lnodeP);
      assert (m_curStudyCardNode_lnodeP);
      
      FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
        lnode_get ((lnode_t*) lnode_get (m_curStudyCardNode_lnodeP));

      /*
       * Skip over Cards that have been moved into the Trash CardBox:
       */
      while (cardInBoxP->cardBoxP->expirationPeriod_hrs == (uint32_t) -1)
        {
          fprintf (stderr, "card is in trash CardBox\n");

          m_curStudyCardNode_lnodeP = 
            list_next (m_studyDynSeqNode_listP, m_curStudyCardNode_lnodeP);
          assert (m_curStudyCardNode_lnodeP);
          cardInBoxP = (FlashCardInBoxType*) 
            lnode_get ((lnode_t*) lnode_get (m_curStudyCardNode_lnodeP));

          if (list_last (m_studyDynSeqNode_listP) == m_curStudyCardNode_lnodeP)
            goto GetNextStudyCard_walk_state_machine; // not all goto are bad
        }

      return cardInBoxP;
    }

  /*
   * Walk the state machine to generate the sequence of study cards:
   */
 GetNextStudyCard_walk_state_machine:

  /* are we in test session?: */
  if (m_test_numCards)
    {
      fprintf (stderr, "Test Card: m_test_curIdx = %d, m_test_numCards = %d\n",
               m_test_curIdx, m_test_numCards);

      /* are we done? */
      m_test_curIdx++;          // m_test_curIdx starts from -1
      if (m_test_curIdx >= (m_test_numCards - 1))
        {
          studyCard_lnodeP = NULL;
          doneStudy_flg = true;
        }
      /* nope, get next card */
      else
        {
          studyCard_lnodeP = m_test_lnodePA [m_test_curIdx];
          assert (studyCard_lnodeP);
        }
    }

  /* nope, we in review or study session: */
  else
    {
      list_t *listP = NULL;
      int numTries_int = 0;
      do {
        numTries_int ++;
        
        switch (m_studySeqState_enum)
          {
          case STUDY_SEQ_STATE_NONE:
            {
              fprintf (stderr, "STUDY_SEQ_STATE_NONE\n");
              m_studySeqState_enum = STUDY_SEQ_STATE_HIGHER_CARDBOXES_EXPIRED;
              
              /* no higher cardboxes at all: */
              if (m_numCardBoxes == 1)
                goto timesUp_lbl;
              
              /* Do we want to review only a particular cardbox? */
              if (m_reviewNthBoxOnlg_idx == -1) // Yes
                {
                  fprintf (stderr, "first of cardbox2\n");
                  listP = (m_cardBoxPA [1])->expiredCards_listP;
                  m_studyCardBox_idx = 1; // at the CardBox\#2
                  studyCard_lnodeP = m_studyCard_lnodePA [1] = 
                    list_first (listP);
                }
              else                  // No, cycle through cardboxes:
                {
                  assert (m_reviewNthBoxOnlg_idx <= m_lastCardBox_idx);
                  listP = (m_cardBoxPA [m_reviewNthBoxOnlg_idx])->
                    expiredCards_listP;
                  m_studyCardBox_idx = m_reviewNthBoxOnlg_idx; // at nth cardbox
                  studyCard_lnodeP = 
                    m_studyCard_lnodePA [m_reviewNthBoxOnlg_idx] = 
                    list_first (listP);
                }
            }
            break;

          case STUDY_SEQ_STATE_HIGHER_CARDBOXES_EXPIRED:
            {
              /* has the time limit reached? */
              assert (m_box1Study_mins > 0);
              if ((curTimeAt_mins - m_startStudyTime_mins) >= 
                  m_higherBoxesReview_mins)
                {
                  fprintf (stderr, "times up studying higher cardboxes\n");
                  goto timesUp_lbl; // not all goto are bad
                }
              
              /* out of cards ? */
              if (unlikely (numTries_int > m_numCardBoxes))
                {
                  fprintf (stderr, "out of higher cardbox card\n");
                  numTries_int = 0; // reset 
                  
                timesUp_lbl:        // not all goto are bad...
                  m_higherBoxesReview_mins = // record the actual time used 
                    curTimeAt_mins - m_startStudyTime_mins;
                  
                  m_studySeqState_enum = STUDY_SEQ_STATE_CARDBOX1_UNVISITED;
                  // TC 2009-08-10 -- Here during the review, demoted cards
                  // ended up in the dontneedvisitcards_listp list, which do
                  // not need to be visited again because all expired cards
                  // are moved into the expired list at the beginning for the
                  // study session.
                  
                  /* move to the cardbox 1's list: */
                  listP = (m_cardBoxPA [0])->expiredCards_listP;
                  m_studyCardBox_idx = 0; // at the CardBox\#1
                  studyCard_lnodeP = m_studyCard_lnodePA [0] = 
                    list_first (listP);
                  break;
                }


              /* Do we want to review only a particular cardbox? */
              if (m_reviewNthBoxOnlg_idx == -1) // No
                {
                  // m_studyCardBox_idx++; // make sure cards are processed...
                  
                  // Stay in the cardbox until certain expired cards have
                  // been drawn:
                  static int drawnCards_int = 0;
                  if (drawnCards_int >= 
                      m_numDrawCards_intA [m_studyCardBox_idx])
                    {
                      drawnCards_int = 0;
                      m_studyCardBox_idx++; // make sure cards are processed...
                    }
                  drawnCards_int ++;
                  // fprintf (stderr, "drawnCards_int is %d\n", drawnCards_int);

                  if (m_studyCardBox_idx == m_numCardBoxes)
                    m_studyCardBox_idx = 1; // cycle back to 2nd CardBox

                  // fprintf (stderr, "--------> m_studyCardBox_idx is %d\n",
                  //          m_studyCardBox_idx);

                  if (m_studyCardBox_idx == m_trashCardBox_idx) // one trash
                    {
                      m_studyCardBox_idx = (m_studyCardBox_idx + 1) % 
                        m_numCardBoxes;
                      fprintf (stderr, "skipping trash box\n");
                      if (m_studyCardBox_idx == 0) // skip 1st cardbox
                        m_studyCardBox_idx ++;
                    }
                  // assert (m_studyCardBox_idx != m_trashCardBox_idx);
                  // fprintf (stderr, "higher cardbox# is %d\n", 
                  //          m_studyCardBox_idx + 1);
                  // TC 2010-08-22 (Sun) -- When there are no higher
                  // boxes at all:
                  if (m_studyCardBox_idx == m_trashCardBox_idx)
                    {
                      studyCard_lnodeP = NULL;
                      break;
                    }


                  /* check to see if cardbox has no expired cards, if so,
                     we reset the drawnCards_int count to make it skip to
                     the next cardbox: */
                  listP = (m_cardBoxPA [m_studyCardBox_idx])->
                    expiredCards_listP;
                  assert (listP);
                  if (list_count (listP) == 0)
                    drawnCards_int = m_numCardBoxes + 1; // make >= max drawn
                }
              else      // Yes, we want to review from only a specific cardbox
                {
                  assert (m_reviewNthBoxOnlg_idx <= m_lastCardBox_idx);
                  m_studyCardBox_idx = m_reviewNthBoxOnlg_idx; // at nth cardbox
                }
              
              /* sanity check: */
              assert ((m_studyCardBox_idx > 0) && 
                      (m_studyCardBox_idx < m_numCardBoxes));
              
              /* switch to cardBox's list: */
              listP = (m_cardBoxPA [m_studyCardBox_idx])->expiredCards_listP;
              assert (listP);
                
              
#ifdef SANITY
              /* sanity check */
              if (m_studyCard_lnodePA [m_studyCardBox_idx]) 
                assert (list_contains 
                        (listP, m_studyCard_lnodePA [m_studyCardBox_idx]));
#endif
              
              /* beginning or end of the list? */
              if (m_studyCard_lnodePA [m_studyCardBox_idx])
                studyCard_lnodeP = m_studyCard_lnodePA [m_studyCardBox_idx] = 
                  list_next (listP, m_studyCard_lnodePA [m_studyCardBox_idx]);
              else
                studyCard_lnodeP = m_studyCard_lnodePA [m_studyCardBox_idx] = 
                  list_first (listP);
            }
            break;
            
          case STUDY_SEQ_STATE_CARDBOX1_UNVISITED:
            {
              fprintf (stderr, "STUDY_SEQ_STATE_CARDBOX1_UNVISITED\n");
              listP = (m_cardBoxPA [0])->expiredCards_listP;
              
#ifdef SANITY
              /* sanity check */
              if (m_studyCard_lnodePA [0])
                assert (list_contains (listP, m_studyCard_lnodePA [0]));
#endif
              
              /* has the time limit reached? */
              assert (m_box1Study_mins > 0);
              if ((curTimeAt_mins - m_startStudyTime_mins) >=
                  (m_higherBoxesReview_mins + m_box1Study_mins))
                {
                  fprintf (stderr, "times up studying higher CardBoxes\n");
                  
                  doneStudy_flg = true; // time's up no more card in study
                  studyCard_lnodeP = NULL;
                  break;
                }
              
              /* beginning or end of the list? */
              if (m_studyCard_lnodePA [0])
                studyCard_lnodeP = m_studyCard_lnodePA [0] = 
                  list_next (listP, m_studyCard_lnodePA [0]);
              else
                studyCard_lnodeP = m_studyCard_lnodePA [0] =  
                  list_first (listP);
            }
            break;
            
          default:
            break;
          }


        fprintf (stderr, "numTries_int is %d, m_numCardBoxes is %d\n",
                 numTries_int, m_numCardBoxes);
        if (unlikely (numTries_int > m_numCardBoxes))
          {
            m_reachedEndOfCards_flg = true; // for saving last card
            doneStudy_flg = true;
          }
        
      } while ((studyCard_lnodeP == NULL) && !doneStudy_flg);
    }


  /*
   * done with study?
   */
  if (doneStudy_flg)
    {
      fprintf (stderr, "all done, no more cards to study\n");
      return NULL;
    }

   
  /*
   * Get the flash card:
   */
  FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
    lnode_get (studyCard_lnodeP);
  fprintf (stderr, "cardInBoxP is %p\n", cardInBoxP);
  assert (cardInBoxP);
  fprintf (stderr, "***** cardInBoxP is %p\n", cardInBoxP);


  /*
   * update the list of studied cards and the current card:
   */
  lnode_t* lnodeP = lnode_create ((void*) studyCard_lnodeP);
  list_append (m_studyDynSeqNode_listP, lnodeP);
  m_curStudyCardNode_lnodeP = lnodeP;


  /*
   * update the visited time of the retrieved card:
   */
  m_lastCardOriVisitedAt_mins = cardInBoxP->visitedAt_mins;
  cardInBoxP->visitedAt_mins = curTimeAt_mins;
  m_lastVisitedAt_mins = curTimeAt_mins; // save the last visitedAt time.


//   fprintf (stderr, "\t\t-- from next got card %d '%s'\n", 
//            cardInBoxP->flashCardP->wordUiq_id,
//            cardInBoxP->flashCardP->front_str);

  return cardInBoxP;
}



FlashCardInBoxType* LeitnerModelClass::
GetPrevCard ()
{
  /*
   * Check to see if the current card has been processed:
   */
  FlashCardInBoxType *curCardInBoxP = GetCurCard ();
  if (curCardInBoxP && (curCardInBoxP->processed_enum == CARD_NOT_PROCESSED))
    {
      fprintf (stderr, "You must process the current card first\n");
      return curCardInBoxP;
    }


  /*
   * Retrieve the previous card from already traversed list:
   */
  if (m_curStudyCardNode_lnodeP &&
      (list_first (m_studyDynSeqNode_listP) != m_curStudyCardNode_lnodeP))
    {
      lnode_t *oriCurStudyCardNode_lnodeP = // in case need to restore
        m_curStudyCardNode_lnodeP;

      m_curStudyCardNode_lnodeP = 
        list_prev (m_studyDynSeqNode_listP, m_curStudyCardNode_lnodeP);
      assert (m_curStudyCardNode_lnodeP);

      FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
        lnode_get ((lnode_t*) lnode_get (m_curStudyCardNode_lnodeP));

//       fprintf (stderr, "\t\tfrom prev <-- got card %d '%s'\n", 
//                cardInBoxP->flashCardP->wordUiq_id,
//                cardInBoxP->flashCardP->front_str);

      /*
       * Skip over Cards that have been moved into the Trash CardBox:
       */
      while (cardInBoxP->cardBoxP->expirationPeriod_hrs == (uint32_t) -1)
        {
          fprintf (stderr, "card is in trash CardBox\n");

          /* first card is also in the Trash CardBox: */
          if (list_first (m_studyDynSeqNode_listP) == m_curStudyCardNode_lnodeP)
            {
              fprintf (stderr, "no more previous none-trash card\n");
              m_curStudyCardNode_lnodeP = // restore the current card
                oriCurStudyCardNode_lnodeP;
              return NULL;
            }

          m_curStudyCardNode_lnodeP = 
            list_prev (m_studyDynSeqNode_listP, m_curStudyCardNode_lnodeP);
          assert (m_curStudyCardNode_lnodeP);
          cardInBoxP = (FlashCardInBoxType*) 
            lnode_get ((lnode_t*) lnode_get (m_curStudyCardNode_lnodeP));
        }

      return cardInBoxP;
    }

  fprintf (stderr, "no more previous card\n");
  return NULL;
}


FlashCardInBoxType* LeitnerModelClass::
GetCurCard ()
{
  if (m_curStudyCardNode_lnodeP == NULL)
    return NULL;

  FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
    lnode_get ((lnode_t*) lnode_get (m_curStudyCardNode_lnodeP));

  return cardInBoxP;
}


void LeitnerModelClass::
Promote ()
{
  /*
   * Check to see if have any study node sequence yet:
   */
  if (m_curStudyCardNode_lnodeP == NULL)
    return;


  /*
   * Get the current card for promotion:
   */
  FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
    lnode_get ((lnode_t*) lnode_get (m_curStudyCardNode_lnodeP));
  // fprintf (stderr, "promoting card %d '%s', pcount=%d\n", 
  //          cardInBoxP->flashCardP->wordUiq_id,
  //          cardInBoxP->flashCardP->front_str,
  //          cardInBoxP->promotionCount);
  
  lnode_t *curStudyCard_lnodeP = (lnode_t*) 
    lnode_get (m_curStudyCardNode_lnodeP);


  /*
   * Calculate which cardbox it's to go in (but skip Trash CardBox): 
   */
  int toBox_idx = cardInBoxP->oriCardBoxIdx + 1;
  if (toBox_idx == m_numCardBoxes)
    toBox_idx --;               // last cardbox reached.
  else if (toBox_idx == m_trashCardBox_idx)
    {
      toBox_idx ++;           // hop over the trash cardbox
      if (toBox_idx == m_numCardBoxes)
        toBox_idx -= 2;               // last cardbox reached.
    }
  assert ((toBox_idx < m_numCardBoxes) && (toBox_idx != m_trashCardBox_idx));


  /*
   * Change previously visited (during to study session) to promote:
   */
  if (cardInBoxP->processed_enum != CARD_NOT_PROCESSED)
    {
      fprintf (stderr, "Promoting previous card\n");

      /*
       * Is card already promoted?  (i.e., not in the CardBox\#1)
       */
      if (cardInBoxP->cardBoxP != m_cardBoxPA [0]) // yep, promoted before
        {
          /* mark card processed and needs saving: */
          cardInBoxP->processed_enum = CARD_PROMOTED; // redundant? 
          m_needSaving_flg = true;

          /* updating the test status array: -- redundant? */
          if (m_test_numCards)
            {
              assert (m_test_curIdx >= 0 && m_test_curIdx < m_test_numCards);
              m_test_statusA [m_test_curIdx] = TEST_YEP;
              m_test_needSavingFlg = true;
            }

          fprintf (stderr, "already promoted, no cardbox changes\n");
          return;               // no cardbox changes
        }

      /*
       * Is the card "marked for promotion" only?
       * (the card is not originally from a higher CardBox)
       */
      if ((cardInBoxP->oriCardBoxIdx == 0) &&
          (cardInBoxP->oriPromotionCount < m_promoteCountThres))
        {
          cardInBoxP->promotionCount = cardInBoxP->oriPromotionCount + 1;
          fprintf (stderr, "*** count = %d not reached promotion yet,\n",
                   cardInBoxP->promotionCount);

          /* mark card processed and needs saving: */
          cardInBoxP->processed_enum = CARD_MARKED_FOR_PROMOTION; // marked
          m_needSaving_flg = true;
          return;
        }
      else
        cardInBoxP->promotionCount = 0; // leaving CardBox\#1, dont need it


      /* 
       * Promotion from previously visited card (to higher CardBoxes):
       */
      list_t *curNotExpd_listP = 
        cardInBoxP->cardBoxP->dontNeedVisitCards_listP;
#ifdef SANITY
      assert (list_contains (curNotExpd_listP, curStudyCard_lnodeP));
#endif
      
      /* remove the card from its current list: */
      list_delete (curNotExpd_listP, curStudyCard_lnodeP);
      
      /* update the card's box pointer to higher CardBox1: */
      cardInBoxP->cardBoxP = m_cardBoxPA [toBox_idx];

      /* mark card processed and needs saving: */
      cardInBoxP->processed_enum = CARD_PROMOTED; // marked
      m_needSaving_flg = true;

      /* updating the test status array: */
      if (m_test_numCards)
        {
          assert (m_test_curIdx >= 0 && m_test_curIdx < m_test_numCards);
          m_test_statusA [m_test_curIdx] = TEST_YEP;
          m_test_needSavingFlg = true;
        }
      
      /* move (append) to the higher CardBox's
         dontneedvisitcards_listp: */
      list_append ((m_cardBoxPA [toBox_idx])->dontNeedVisitCards_listP,
                   curStudyCard_lnodeP);
      
      return;
    }



  /* 
   * Promotion from the most recent card:
   */
  list_t *card_listP = cardInBoxP->cardBoxP->expiredCards_listP;

  /* card is in CardBox\#1?, which list?: */
  if (cardInBoxP->cardBoxP == m_cardBoxPA [0])
    {
      if (cardInBoxP->promotionCount < m_promoteCountThres)
        {
          cardInBoxP->promotionCount = cardInBoxP->oriPromotionCount + 1;
          fprintf (stderr, "*** not reached promotion yet, same as demotion\n");
          Demote ();

          /* mark card processed and needs saving: */
          cardInBoxP->processed_enum = CARD_MARKED_FOR_PROMOTION; // marked
          m_needSaving_flg = true;
          return;
        }

//       if (m_studySeqState_enum == STUDY_SEQ_STATE_CARDBOX1_VISITED)
//         card_listP = cardInBoxP->cardBoxP->dontNeedVisitCards_listP;
      assert (m_studySeqState_enum != STUDY_SEQ_STATE_CARDBOX1_VISITED);
      cardInBoxP->promotionCount = 0; // leaving CardBox#1, don't need it
    }


  /* update the study sequence before deletion from the list: */
#ifdef SANITY
  assert (list_contains (card_listP, curStudyCard_lnodeP)); // sanity
#endif
  m_studyCard_lnodePA [m_studyCardBox_idx] = 
    list_prev (card_listP, curStudyCard_lnodeP);

  /* remove the card from its current list: */
  list_delete (card_listP, curStudyCard_lnodeP);
  
  /* update the card's box pointer to higher CardBox1: */
  assert (toBox_idx > 0);
  cardInBoxP->cardBoxP = m_cardBoxPA [toBox_idx];

  /* mark card processed and needs saving: */
  cardInBoxP->processed_enum = CARD_PROMOTED; // mark promoted
  m_needSaving_flg = true;
  
  /* updating the test status array: */
  if (m_test_numCards)
    {
      assert (m_test_curIdx >= 0 && m_test_curIdx < m_test_numCards);
      m_test_statusA [m_test_curIdx] = TEST_YEP;
      m_test_needSavingFlg = true;
    }

  /* move (append) to the higher CardBox's dontneedvisitcards_listp: */
  list_append ((m_cardBoxPA [toBox_idx])->dontNeedVisitCards_listP,
               curStudyCard_lnodeP);
}




void LeitnerModelClass::
Demote ()
{
  /*
   * Check to see if have any study node sequence yet:
   */
  if (m_curStudyCardNode_lnodeP == NULL)
    return;

  /*
   * Get the current card for demotion:
   */
  lnode_t *curStudyCard_lnodeP = (lnode_t*) 
    lnode_get (m_curStudyCardNode_lnodeP);

  FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
    lnode_get (curStudyCard_lnodeP);
  // fprintf (stderr, "demoting card %d %s\n", 
  //          cardInBoxP->flashCardP->wordUiq_id,
  //          cardInBoxP->flashCardP->front_str);

  /*
   * Change previously visited (during to study session) to demote:
   */
  if (cardInBoxP->processed_enum != CARD_NOT_PROCESSED)
    {
      fprintf (stderr, "Demoting previous card\n");

      /*
       * Is card already demoted?  (i.e., in the CardBox\#1)
       */
      if (cardInBoxP->cardBoxP == m_cardBoxPA [0]) // yes, already demoted
        {
          /* undo promotion count: */
          if (cardInBoxP->oriCardBoxIdx == 0)
            cardInBoxP->promotionCount = cardInBoxP->oriPromotionCount;
          else
            cardInBoxP->promotionCount = m_promoteCountThres;

          /* mark card processed and needs saving: */
          cardInBoxP->processed_enum = CARD_DEMOTED; // mark demoted
          m_needSaving_flg = true;


          /* updating the test status array: */
          if (m_test_numCards)
            {
              assert (m_test_curIdx >= 0 && m_test_curIdx < m_test_numCards);
              m_test_statusA [m_test_curIdx] = TEST_NOPE;
              m_test_needSavingFlg = true;
            }

          fprintf (stderr, "already demoted, no change, pCount reset  = %d\n", 
                   cardInBoxP->promotionCount);
          return;                 // no change
        }

      /* 
       * Demotion from previously visited card (now in higher cardbox):
       */
      /* Calculate which cardbox it is in now: */
      int nowInBox_idx;
      for (nowInBox_idx = cardInBoxP->oriCardBoxIdx; 
           nowInBox_idx < m_numCardBoxes; nowInBox_idx ++)
        {
          if (cardInBoxP->cardBoxP == m_cardBoxPA [nowInBox_idx])
            break;
        }
      assert (nowInBox_idx < m_numCardBoxes);

      /* set promotion count to max if was from higher cardbox: */
      cardInBoxP->promotionCount = m_promoteCountThres;

      /* remove the card from its current list: */
      list_t *curNotExpd_listP = 
        m_cardBoxPA [nowInBox_idx]->dontNeedVisitCards_listP;
#ifdef SANITY
      assert (list_contains (curNotExpd_listP, curStudyCard_lnodeP));
#endif
      list_delete (curNotExpd_listP, curStudyCard_lnodeP);

      /* update the card's box pointer to CardBox\#1: */
      cardInBoxP->cardBoxP = m_cardBoxPA [0];

      /* mark card processed and needs saving: */
      cardInBoxP->processed_enum = CARD_DEMOTED; // mark demoted
      m_needSaving_flg = true;

      /* updating the test status array: */
      if (m_test_numCards)
        {
          assert (m_test_curIdx >= 0 && m_test_curIdx < m_test_numCards);
          m_test_statusA [m_test_curIdx] = TEST_NOPE;
          m_test_needSavingFlg = true;
        }

      /* move prepend to the CardBox\#1's dontneedvisitcards_listp: */
      list_prepend ((m_cardBoxPA [0])->dontNeedVisitCards_listP,
                    curStudyCard_lnodeP);
      return;
    }


  /* 
   * Demotion from the most recent card:
   */
  list_t *card_listP = cardInBoxP->cardBoxP->expiredCards_listP;
  assert (m_studySeqState_enum != STUDY_SEQ_STATE_CARDBOX1_VISITED);

  /* update the study sequence before deletion from the list: */
#ifdef SANITY
  assert (list_contains (card_listP, curStudyCard_lnodeP));
#endif
  m_studyCard_lnodePA [m_studyCardBox_idx] = 
    list_prev (card_listP, curStudyCard_lnodeP);

  /* remove the card from its current list: */
  list_delete (card_listP, curStudyCard_lnodeP);
  
  /* update the card's box pointer to CardBox\#1: */
  cardInBoxP->cardBoxP = m_cardBoxPA [0];

  /* mark card processed and needs saving: */
  cardInBoxP->processed_enum = CARD_DEMOTED; // mark demoted
  m_needSaving_flg = true;

  /* updating the test status array: */
  if (m_test_numCards)
    {
      assert (m_test_curIdx >= 0 && m_test_curIdx < m_test_numCards);
      m_test_statusA [m_test_curIdx] = TEST_NOPE;
      m_test_needSavingFlg = true;
    }

  /* set promotion count to max if was from higher cardbox: */
  if (cardInBoxP->oriCardBoxIdx > 0)
    cardInBoxP->promotionCount = m_promoteCountThres;

  /* move (prepend) to the CardBox\#1's dontneedvisitcards_listp: */
  list_prepend ((m_cardBoxPA [0])->dontNeedVisitCards_listP,
                curStudyCard_lnodeP);

  // fprintf (stderr, "@#@#@#@#@#@##@@#@ %p is in %p\n",
  //          curStudyCard_lnodeP, (m_cardBoxPA [0])->dontNeedVisitCards_listP);
  // fprintf (stderr, "dont need save is now %p = %d cards\n",
  //          (m_cardBoxPA [0])->dontNeedVisitCards_listP,
  //          (int) list_count ((m_cardBoxPA [0])->dontNeedVisitCards_listP));
}



void LeitnerModelClass::
Trash ()
{
  /*
   * Check to see if have any study node sequence yet:
   */
  if (m_curStudyCardNode_lnodeP == NULL)
    return;

  /*
   *  Is there a Trash CardBox?
   */
  if (m_trashCardBox_idx == -1)
    {
      fprintf (stderr, "There is no Trash CardBox yet\n");
      return;
    }
  
  /*
   * Get the current card for trashing:
   */
  lnode_t *curStudyCard_lnodeP = (lnode_t*) 
    lnode_get (m_curStudyCardNode_lnodeP);

  FlashCardInBoxType* cardInBoxP = (FlashCardInBoxType*) 
    lnode_get (curStudyCard_lnodeP);
  fprintf (stderr, "trashing card %d %s\n", 
           cardInBoxP->flashCardP->wordUiq_id,
           cardInBoxP->flashCardP->front_str);
  
  /*
   * Change previously visited/processed (during to study session):
   */
  if (cardInBoxP->processed_enum != CARD_NOT_PROCESSED)
    {
      if (unlikely (cardInBoxP->processed_enum == CARD_TRASHED))
        {
          fprintf (stderr, "card already trashed ??\n");
          return;
        }
      else
        /*
         * Undo the previous process to original state:
         */
        {
          fprintf (stderr, "card was previously process as %d\n",
                   cardInBoxP->processed_enum);
          int nowInBox_idx;
          for (nowInBox_idx = 0; nowInBox_idx < m_numCardBoxes; nowInBox_idx++)
            {
              if (cardInBoxP->cardBoxP == m_cardBoxPA [nowInBox_idx])
                break;
            }
          assert (nowInBox_idx < m_numCardBoxes);

          /* restore pCount: */
          cardInBoxP->promotionCount = cardInBoxP->oriPromotionCount;

          /* restore processed status: */
          cardInBoxP->processed_enum = CARD_NOT_PROCESSED;

          /* remove the card from its current list: */
          list_t *curNotExpd_listP = 
            m_cardBoxPA [nowInBox_idx]->dontNeedVisitCards_listP;
#ifdef SANITY
          assert (list_contains (curNotExpd_listP, curStudyCard_lnodeP));
#endif
          list_delete (curNotExpd_listP, curStudyCard_lnodeP);

          /* update the card's box pointer to original cardbox : */
          cardInBoxP->cardBoxP = m_cardBoxPA [cardInBoxP->oriCardBoxIdx];

          /* move back to the original CardBox's expiredCards_list: */
          list_prepend ((cardInBoxP->cardBoxP)->expiredCards_listP,
                        curStudyCard_lnodeP);

        }
    }


  /* 
   * Forgetting from the most recent card:
   */
  list_t *card_listP = cardInBoxP->cardBoxP->expiredCards_listP;

//   /* card is in CardBox\#1?, which list?: */
//   if ((cardInBoxP->cardBoxP == m_cardBoxPA [0])
//       && (m_studySeqState_enum == STUDY_SEQ_STATE_CARDBOX1_VISITED))
//     card_listP = cardInBoxP->cardBoxP->dontNeedVisitCards_listP;
  assert (m_studySeqState_enum != STUDY_SEQ_STATE_CARDBOX1_VISITED);

  /* update the study sequence before deletion from the list: */
#ifdef SANITY
  assert (list_contains (card_listP, curStudyCard_lnodeP));
#endif
  m_studyCard_lnodePA [m_studyCardBox_idx] = 
    list_prev (card_listP, curStudyCard_lnodeP);

  /* remove the card from its current list: */
  list_delete (card_listP, curStudyCard_lnodeP);

  /* update the card's box pointer to Trash CardBox: */
  assert (m_trashCardBox_idx > 0);
  cardInBoxP->cardBoxP = m_cardBoxPA [m_trashCardBox_idx];

  /* mark card processed and needs saving: */
  cardInBoxP->processed_enum = CARD_TRASHED; // mark trashed
  m_needSaving_flg = true;

  /* move (prepend) to the Trash CardBox's dontneedvisitcards_listp: */
  list_prepend ((m_cardBoxPA [m_trashCardBox_idx])->dontNeedVisitCards_listP,
                curStudyCard_lnodeP);
  fprintf (stderr, "dont need save is now %p = %d cards\n",
           (m_cardBoxPA [m_trashCardBox_idx])->dontNeedVisitCards_listP,
           (int) list_count ((m_cardBoxPA [m_trashCardBox_idx])->
                             dontNeedVisitCards_listP));
}



bool LeitnerModelClass::
GetNumCards (int        *out_numTotalActiveCards_intP, // cards visited
             int        *out_numCb1Visited_int,
             int        *out_numCb1Unvisited_int,
             int        *out_numHcbsDontNeedVisite_int,
             int        *out_numHcbsExpired_int,
             int        *out_numNeverVistied_int,
             int        *out_numImaturePcounts_int,
             int        *out_numCardBoxes_int,
             const int **out_numExpireds_intAP)
{
  if (m_opened_flg == false)
    {
      if (out_numTotalActiveCards_intP)
        *out_numTotalActiveCards_intP = 0;

      if (out_numCb1Visited_int)
        *out_numCb1Visited_int = 0;
      
      if (out_numCb1Unvisited_int)
        *out_numCb1Unvisited_int = 0;

      if (out_numHcbsDontNeedVisite_int)
        *out_numHcbsDontNeedVisite_int = 0;

      if (out_numHcbsExpired_int)
        *out_numHcbsExpired_int = 0;
      
      if (out_numNeverVistied_int)
        *out_numNeverVistied_int = 0;
      
      if (out_numImaturePcounts_int)
        *out_numImaturePcounts_int = 0;

      if (out_numCardBoxes_int)
        *out_numCardBoxes_int = 0;
    
      fprintf (stderr, "cardcabinetfile not opened yet\n");
      return false;
    }

  if (out_numTotalActiveCards_intP)
    {
      *out_numTotalActiveCards_intP = 
        list_count ((m_cardBoxPA [0])->dontNeedVisitCards_listP) + 
        list_count ((m_cardBoxPA [0])->expiredCards_listP);

      for (int idx = 1; idx < m_numCardBoxes; idx++)
        {
          /* skip trash cardbox: */
          if (idx == m_trashCardBox_idx)
            continue;

          *out_numTotalActiveCards_intP +=
            list_count ((m_cardBoxPA [idx])->dontNeedVisitCards_listP);
      
          *out_numTotalActiveCards_intP +=
            list_count ((m_cardBoxPA [idx])->expiredCards_listP);
        }
    }


  if (out_numCb1Visited_int)
    *out_numCb1Visited_int = 
      list_count ((m_cardBoxPA [0])->dontNeedVisitCards_listP);

  if (out_numCb1Unvisited_int)
    *out_numCb1Unvisited_int = 
      list_count ((m_cardBoxPA [0])->expiredCards_listP);

  if (out_numHcbsDontNeedVisite_int)
    {
      *out_numHcbsDontNeedVisite_int = 0;
      for (int idx = 1; idx < m_numCardBoxes; idx++)
        *out_numHcbsDontNeedVisite_int +=
          list_count ((m_cardBoxPA [idx])->dontNeedVisitCards_listP);
    }

  if (out_numHcbsExpired_int)
    {
      *out_numHcbsExpired_int = 0;
      fprintf (stderr, "m_numCardBoxes is %d\n", m_numCardBoxes);
      for (int idx = 1; idx < m_numCardBoxes; idx++)
        {
          /* skip trash cardbox: */
          if (idx == m_trashCardBox_idx)
            continue;

          *out_numHcbsExpired_int +=
            list_count ((m_cardBoxPA [idx])->expiredCards_listP);
          fprintf (stderr, "cardbox %d has %d expired cards\n",
                   idx, (int) 
                   list_count ((m_cardBoxPA [idx])->expiredCards_listP));
        }
    }

  if (out_numNeverVistied_int)
    *out_numNeverVistied_int = m_neverVisitedCard_int;

  /* count number of cards that does not have enough pCount: */
  if (out_numImaturePcounts_int)
    {
      list_t *cards_listP = m_cardBoxPA [0]->expiredCards_listP;
      lnode_t *lnodeP = list_first (cards_listP);
      *out_numImaturePcounts_int = 0;
      int numCards = list_count (cards_listP);
      FlashCardInBoxType *cardInBoxP;
      for (int idx = 0; idx < numCards; idx++)
        {
          cardInBoxP = (FlashCardInBoxType*) lnode_get (lnodeP);
          if (cardInBoxP->promotionCount < m_promoteCountThres)
            *out_numImaturePcounts_int ++;
          lnodeP = list_next (cards_listP, lnodeP);
        }
    }

  if (out_numCardBoxes_int)
    *out_numCardBoxes_int = m_numCardBoxes;

  if (out_numExpireds_intAP)
    {
      for (int idx = 0; idx < m_numCardBoxes; idx++)
        {
          m_numExpireds_intA [idx] = 
            list_count ((m_cardBoxPA [idx])->expiredCards_listP);
        }
      *out_numExpireds_intAP = m_numExpireds_intA;
    }
  return true;
}


static void 
PopulateCardArray (list_t               *cards_listP,
                   int                   numCards,
                   FlashCardInBoxType ***cardsPAP)
{
      
  FlashCardInBoxType *cardInBoxP;
  lnode_t *lnodeP = list_first (cards_listP);

  for (int idx = 0; idx < numCards; idx++)
    {
      cardInBoxP = (FlashCardInBoxType*) lnode_get (lnodeP);
      (*cardsPAP)[idx] = cardInBoxP;
      lnodeP = list_next (cards_listP, lnodeP);
    }
}


bool LeitnerModelClass::
GetAllCardsInBox (int                   in_boxNum, // start from 0
                  int                  *out_numDontNeedVisit_intP,
                  FlashCardInBoxType ***out_dontNeedVisit_cardsPAP,
                  int                  *out_numExpried_intP,
                  FlashCardInBoxType ***out_expired_cardsPAP)
{
  /*
   * Error checking:
   */
  if ((in_boxNum < 0) || (in_boxNum >= m_numCardBoxes))
    {
      fprintf (stderr, "in_boxNum invalid\n");
      return false;
    }


  /*
   * Get CardBox info:
   */
  CardBoxType *cardBoxP = m_cardBoxPA [in_boxNum];
  int numDontNeedVisitCards = 
    (int) list_count (cardBoxP->dontNeedVisitCards_listP);
  int numExpiredCards = 
    (int) list_count (cardBoxP->expiredCards_listP);
  if (out_numDontNeedVisit_intP)
    *out_numDontNeedVisit_intP = numDontNeedVisitCards;
  if (out_numExpried_intP)
    *out_numExpried_intP = numExpiredCards;
  

  /*
   * re-allocate if necessary:
   */
  Malloc (m_dontNeedVisit_cardsPA, FlashCardInBoxType**, 
          numDontNeedVisitCards * sizeof (FlashCardInBoxType*));
  Malloc (m_expired_cardsPA, FlashCardInBoxType**, 
          numExpiredCards * sizeof (FlashCardInBoxType*));


  /* use user storage */
  UseUserStorage (out_dontNeedVisit_cardsPAP,
                  FlashCardInBoxType**,
                  numDontNeedVisitCards * sizeof (FlashCardInBoxType*));
  UseUserStorage (out_expired_cardsPAP,
                  FlashCardInBoxType**,
                  numExpiredCards * sizeof (FlashCardInBoxType*));

  /* use built-in storage */
  UseBuiltInStorage (out_dontNeedVisit_cardsPAP,
                     m_dontNeedVisit_cardsPA);
  UseBuiltInStorage (out_expired_cardsPAP,
                     m_expired_cardsPA);
  
  
  /*
   * populate output array:
   */
  if (out_dontNeedVisit_cardsPAP)
    PopulateCardArray (cardBoxP->dontNeedVisitCards_listP,
                       numDontNeedVisitCards,
                       out_dontNeedVisit_cardsPAP);

  if (out_expired_cardsPAP)
    PopulateCardArray (cardBoxP->expiredCards_listP,
                       numExpiredCards,
                       out_expired_cardsPAP);

  return true;
}


uint32_t LeitnerModelClass::
GetRemainStudyTime ()
{
  uint32_t curTime_mins = GetMinSinceEpoch ();
  // fprintf (stderr, "curTime_mins is %d\n", curTime_mins);
  // fprintf (stderr, "m_startStudyTime_mins %d\n", m_startStudyTime_mins);
  // fprintf (stderr, "elsps mins is %d\n", 
  //          curTime_mins - m_startStudyTime_mins);


  int remainingTime_min = (m_box1Study_mins + m_higherBoxesReview_mins) - 
    (curTime_mins - m_startStudyTime_mins);
  // fprintf (stderr, "m_box1Study_mins = %d, m_higherBoxesReview_mins = %d\n",
  //          m_box1Study_mins, m_higherBoxesReview_mins); 
  // fprintf (stderr, "remaining min is %d\n", remainingTime_min);

  /* if the user idles for too long, remaining time could become negative. */
  if (remainingTime_min < 0)
    remainingTime_min = 0;

  return remainingTime_min;
}


bool LeitnerModelClass::
ResumeTimer ()
{
  /*
   * Do not resume session timer if last visited was within pause timeout:
   */
  uint32_t curTime_mins = GetMinSinceEpoch ();
  if (likely ((curTime_mins - m_lastVisitedAt_mins) < m_puaseTimeout_uint))
    return false;


  /*
   * Resume session timer and state:
   */
  int shiftToLastVisited = curTime_mins - m_lastVisitedAt_mins;
  m_lastVisitedAt_mins += shiftToLastVisited;
  m_startStudyTime_mins += shiftToLastVisited;

  fprintf (stderr, "set m_startStudyTime_mins to %d, remaining time is %d\n",
           m_startStudyTime_mins, GetRemainStudyTime ());

  /*
   * Reset pause timeout to default if was set to 0 (immediate pause):
   */
  if (m_puaseTimeout_uint == 0)
    m_puaseTimeout_uint = DEFAULT_PAUSE_TIMEOUT_MIN;

  return true;
}


void LeitnerModelClass::
SetPauseTimeOut (unsigned int in_timeoutMin_uint)
{
  m_puaseTimeout_uint = in_timeoutMin_uint;  

}



void LeitnerModelClass::
PrintCardInBoxInfo (const FlashCardInBoxType *in_cardInBoxP)
{
  fprintf (stderr, "\n");
  fprintf (stderr, "=================== CARD INFO   =======\n");
  fprintf (stderr, "%p: card_id %d front = '%s', pcount=%d\n", 
           in_cardInBoxP,
           in_cardInBoxP->flashCardP->wordUiq_id,
           in_cardInBoxP->flashCardP->front_str,
           in_cardInBoxP->promotionCount);
  int cardBox_int;
  for (cardBox_int = 0; cardBox_int < m_numCardBoxes; cardBox_int++)
    if (in_cardInBoxP->cardBoxP == m_cardBoxPA [cardBox_int])
      break;
  assert (cardBox_int < m_numCardBoxes);
  fprintf (stderr, "cardbox = %d\n", cardBox_int);
  fprintf (stderr, "\n");
  fprintf (stderr, "\n");
}



void LeitnerModelClass::
SetReviewLastBoxOnly (bool in_lastBoxOnly_flg)
{
  /* Note: Must call this function again after a new cabinet file is
     opened. */
  // m_reviewLastBoxOnlg_flg = in_lastBoxOnly_flg;
  // fprintf (stderr, "set m_reviewLastBoxOnlg_flg to %d\n",
  //          m_reviewLastBoxOnlg_flg);

  if (in_lastBoxOnly_flg)
    m_reviewNthBoxOnlg_idx = m_lastCardBox_idx;
  else
    m_reviewNthBoxOnlg_idx = -1;
}


bool LeitnerModelClass::
SetReview2ndBoxOnly (bool in_2ndBoxOnly_flg)
{
  // m_reviewLastBoxOnlg_flg = in_lastBoxOnly_flg;
  // fprintf (stderr, "set m_reviewLastBoxOnlg_flg to %d\n",
  //          m_reviewLastBoxOnlg_flg);

  /* do we even have 2nd cardbox? */
  if (m_lastCardBox_idx < 1)
    return false;

  if (in_2ndBoxOnly_flg)
    m_reviewNthBoxOnlg_idx = 1; 
  else
    m_reviewNthBoxOnlg_idx = -1;

  return true;
}


void* 
CreateStorage ()
{
  int *storageP = NULL;
  Malloc (storageP, int*, 1);
  return storageP;
}

void 
DestroyStorage (void *storageP)
{
  Free (storageP);
}

void
ReAllocateStoreage (void *io_storagePP,
                    int   in_bytes)
{
  Malloc (*((void**)io_storagePP), void*, in_bytes);
}



bool LeitnerModelClass::
LoadTestSession (const char *in_testFile_str)
{
  if (m_opened_flg == false)
    {
      fprintf (stderr, "must call open() first\n");
      return false;
    }

  /*
   * Allocate storage for test cards:
   */
  int numActiveCards_int;
  GetNumCards (&numActiveCards_int, NULL, NULL, NULL, 
               NULL, NULL, NULL, NULL, NULL);
  fprintf (stderr, "numActiveCards_int is %d\n", numActiveCards_int);
  assert (m_test_lnodePA == NULL);
  m_test_lnodePA = (lnode_t**) 
    malloc (sizeof (lnode_t*) * numActiveCards_int);
  assert (m_test_statusA == NULL);
  m_test_statusA = (uint8_t*) 
    malloc (sizeof (uint8_t) * numActiveCards_int);


  /*
   * Create a new test session:
   */
  FILE *test_fileP;
  if (in_testFile_str == NULL)  // use default test file
    {
      char *testFile_str = strdup (m_cardBoxCabinetFile_str);
      char *fileExt_chrP = strstr (testFile_str, ".cbx");
      assert (fileExt_chrP);
      if (fileExt_chrP == NULL)
        {
          fprintf (stderr, "could not open default test file for saving\n");
          return false;
        }

      sprintf (fileExt_chrP, ".tst"); // overwrite file extension
      test_fileP = fopen (testFile_str, "rb");
      free (testFile_str);
    }
  else
    test_fileP = fopen (in_testFile_str, "rb");
  if (test_fileP == NULL)
    {
      /* create a temporary test cards array: */
      lnode_t *tmpTest_lnodeA [numActiveCards_int]; // exists locally only
      int tmp_idx = 0;
      for (int box_idx = 0; box_idx < m_numCardBoxes; box_idx++)
        {
          if (unlikely (box_idx == m_trashCardBox_idx)) // skip trash card box
            continue;

          list_t *cards_listP = m_cardBoxPA [box_idx]->dontNeedVisitCards_listP;
          lnode_t *lnodeP = list_first (cards_listP);
          while (lnodeP)
            {
              tmpTest_lnodeA [tmp_idx++] = lnodeP;
              assert (tmp_idx <= numActiveCards_int);
              lnodeP = list_next (cards_listP, lnodeP);
            }

          cards_listP = m_cardBoxPA [box_idx]->expiredCards_listP;
          lnodeP = list_first (cards_listP);
          while (lnodeP)
            {
              tmpTest_lnodeA [tmp_idx++] = lnodeP;
              assert (tmp_idx <= numActiveCards_int);
              lnodeP = list_next (cards_listP, lnodeP);
            }
        }

      /*
       * Create permutation of indexes:
       */
      int permuteIdxA [numActiveCards_int];
      for (int idx = 0; idx < numActiveCards_int; idx++)
        permuteIdxA [idx] = idx;
      CreateKnuthShuffle (permuteIdxA, numActiveCards_int);

      /*
       * Randomize the test cards array:
       */
      for (int idx = 0; idx < numActiveCards_int; idx++)
        {
          m_test_lnodePA [idx] = tmpTest_lnodeA [permuteIdxA [idx]];
          assert (m_test_lnodePA [idx]);
        }
      
      /*
       * Reset the test status array:
       */
      memset (m_test_statusA, TEST_NOT_VISITED, 
              sizeof (uint8_t) * numActiveCards_int);
    }

  /*
   * Load from test file:  Format is as follows:
   *
   *file_id="0" card_id="_2650" status="yep"
   *file_id="0" card_id="_2651" status="not_visited"
   */
  else
    {
      const int maxLen = 80;    // 80 should be enough for everybody
      char line_str [maxLen];
      int numFields;

      int      fileId_int;
      uint32_t cardId_u32;
      char     status_str [maxLen];

      assert (m_test_curIdx == -1);
      int array_idx = 0;
      while (fgets (line_str, maxLen, test_fileP))
        {
          /* skip blank lines: */
          if (line_str [0] == '\n') 
            continue;

          /* read in fields: */
          numFields = sscanf (line_str, 
                              "file_id=\"%d\" card_id=\"_%d\" status=\"%[^\"]",
                              &fileId_int, &cardId_u32, status_str);
          if (unlikely (numFields != 3))
            {
              fprintf (stderr, "line read: %s\n", line_str);
              fprintf (stderr, "test file is corrupted?\n\n");

              goto fileCorrupt_lbl;
            }

          /* locate card: */
          FlashCardInBoxType dummyCardInBox;
          dummyCardInBox.file_id = fileId_int;
          lnode_t dummy_lnode;
          lnode_init (&dummy_lnode, &dummyCardInBox);
          dnode_t *dnodeP = dict_lookup (m_cbCabinet_dictP, (void*) cardId_u32, 
                                         &dummy_lnode);
          if (unlikely (dnodeP == NULL))
            {
              fprintf (stderr, "could not find card id=%d, file id=%d\n",
                       cardId_u32, fileId_int);
              goto fileCorrupt_lbl;
            }

          lnode_t *lnodeP = (lnode_t*) dnode_get (dnodeP);
          FlashCardInBoxType *cardInBoxP = 
            (FlashCardInBoxType*) lnode_get (lnodeP);
          
          /* skip trash cards (may get introduced later even if loaded
             from supposedly thrash-free test file) */
          if (cardInBoxP->oriCardBoxIdx == -1)
            continue;

          /* populate the array: */
          m_test_lnodePA [array_idx] = lnodeP;
          if (strcmp (status_str, "not_visited") == 0)            
            m_test_statusA [array_idx] = TEST_NOT_VISITED;
          else if (strcmp (status_str, "nope") == 0)
            {
              m_test_statusA [array_idx] = TEST_NOPE;
              m_test_curIdx = array_idx; // skip visited
            }
          else if (strcmp (status_str, "yep") == 0)
            {
              m_test_statusA [array_idx] = TEST_YEP;
              m_test_curIdx = array_idx; // skip visited
            }
          else
            {
              fprintf (stderr, "test file status is corrupted? %s\n\n",
                       status_str);
              goto fileCorrupt_lbl;
            }

          array_idx ++;
          if (unlikely (array_idx > numActiveCards_int))
            {
              fprintf (stderr, "too many cards (%d) in test file ?\n\n",
                       array_idx);

              goto fileCorrupt_lbl;
            }
        }
      assert (m_test_curIdx != (numActiveCards_int - 1)); // some non-visited

      if (unlikely (array_idx != numActiveCards_int))
        {
          fprintf (stderr, "number of test cards do not match\n");
          goto fileCorrupt_lbl;
        }

      fclose (test_fileP);      // success
    }

  fprintf (stderr, "loaded test session, m_test_curIdx is %d\n", m_test_curIdx);
  m_test_numCards = numActiveCards_int; // indicate test session
  return true;                  // success

 fileCorrupt_lbl:
  fclose (test_fileP);
  return false;
}


const FlashCardType* LeitnerModelClass::
GetRandomCard (const FlashCardType *in_flashCardP)
{
  fakeUseParam (in_flashCardP);


  return NULL;
}
