/**
 * @file leitnerSysModel.hh
 *
 *  @remark ~ The back-end interface of the Leitner system.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright 2009 Tommy Chang.  This file is part of gCardSys.
 *
 *  gCardSys is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  gCardSys is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gCardSys.  If not, see <http://www.gnu.org/licenses/>.
 *  `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-05-31 
 **/


#ifndef LEITNERSYSMODEL_HH
#define LEITNERSYSMODEL_HH
/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include "flashCard.hh"
#include "list.h"
#include "dict.h"
#include "lookupTable.hh"

/*-----------------*
 * Macro Constants *
 \*---------------*/


/*-----------------*
 * Data Structures *
 \*---------------*/

typedef enum CardProcessEnum {
  CARD_NOT_PROCESSED = 0,
  CARD_PROMOTED,
  CARD_MARKED_FOR_PROMOTION,
  CARD_DEMOTED,
  CARD_TRASHED,
} CardProcessEnum;

typedef enum StudySeqStateEnum {
  STUDY_SEQ_STATE_NONE = 0,
  STUDY_SEQ_STATE_CARDBOX1_VISITED,   // "aka don't need"
  STUDY_SEQ_STATE_CARDBOX1_UNVISITED, // "aka expired"
  STUDY_SEQ_STATE_HIGHER_CARDBOXES_EXPIRED,
} StudySeqStateEnum;


typedef enum CabinetXmlStateEnum {
  CABINET_XML_STATE_NONE = 0,
  CABINET_XML_STATE_CABINET,
  CABINET_XML_STATE_DECKFILE,
  CABINET_XML_STATE_CARDBOX,
} CabinetXmlStateEnum;


typedef struct CardBoxType {
  list_t   *dontNeedVisitCards_listP; // same as unfamiliar for CardBox#1
  list_t   *expiredCards_listP;      // need immediate attention
  uint32_t  expirationPeriod_hrs;    // expiration period in hours
} CardBoxType;


typedef struct FlashCardInBoxType {
  const FlashCardType *flashCardP;
  int                  file_id;
  int                  visitedAt_mins; // in minutes since 970 1/1/ 00:00am
  int16_t              promotionCount;  // used by cards in CardBox\#1
  int16_t              oriPromotionCount;  // used by cards in CardBox\#1
  int                  oriCardBoxIdx;   // original cardbox after loading
  CardProcessEnum      processed_enum; // card promoted or demoted or none?
  const CardBoxType   *cardBoxP;       // the cardbox the card is in
  FlashCardClass      *deckObjP;       // the deck the card was from
} FlashCardInBoxType;


typedef enum MoveModeEnum {
  MOVE_PREPEND = 0,
  MOVE_APPEND,
} MoveModeEnum;


typedef enum TestStatusEnum {
  TEST_NOT_VISITED = 0,
  TEST_NOPE,
  TEST_YEP,
} TestStatusEnum;


class LeitnerModelClass
{
public:
  LeitnerModelClass ();
  LeitnerModelClass (const LeitnerModelClass &); // avoid compiler warning
  void operator =(const LeitnerModelClass &); // avoid compiler warning
  ~LeitnerModelClass ();

  bool Open (const char *in_cardBoxCabinetFile_str);             
  bool Save (const char *in_cabinetFile_str,
             bool        in_exportDkf_flg,
             bool        in_isClosing_flg);
  bool ExportToDeck (const char *in_deckFile_str);
  bool NeedSave ();
  void Close ();

  FlashCardInBoxType* GetNextCard ();
  FlashCardInBoxType* GetPrevCard ();
  FlashCardInBoxType* GetCurCard ();

  void Promote ();
  void Demote ();
  void Trash ();

  uint32_t GetStudyStartTime () {return m_startStudyTime_mins;};
  uint32_t GetRemainStudyTime ();
  bool     GetNumCards (int        *out_numTotalActiveCards_int,
                        int        *out_numCb1Visited_int,
                        int        *out_numCb1Unvisited_int,
                        int        *out_numHcbsDontNeedVisite_int,
                        int        *out_numHcbsExpired_int,
                        int        *out_numNeverVistied_int,
                        int        *out_numImaturePcounts_int,
                        int        *out_numCardBoxes_int,
                        const int **out_numExpireds_intAP);

  const char *GetCardBoxCabinetDir () {return m_cardBoxCabinetDirOnly_str;};

  bool IsReviewMode () {return m_studyCardBox_idx > 0;};
  void SetReviewLastBoxOnly (bool in_lastBoxOnly_flg);
  bool SetReview2ndBoxOnly (bool in_2ndBoxOnly_flg);
  void SetPauseTimeOut (unsigned int in_timeoutMin_uint);
  bool LoadTestSession (const char *in_testFile_str);
  const FlashCardType* GetRandomCard (const FlashCardType *flashCardP);

  void PrintCardInBoxInfo (const FlashCardInBoxType *in_cardInBoxP);

private:
  void ProcessXmlNode (xmlTextReaderPtr in_readerP);
  bool LoadDeckFiles ();
  void UnloadDeckFiles ();
  static int LoadFlashCards (const void *data, 
                             const void *user);

  void InsertCardIntoBox (int                  in_file_id, 
                          FlashCardClass      *in_deckP,
                          const FlashCardType* in_flashCardP,
                          uint32_t             in_visitedAt_mins,
                          uint32_t             in_pCount,
                          bool                 in_visited_flg);

  void InsertAllCardIntoBox (int      in_file_id,
                             uint32_t in_visitedAt_mins,
                             uint32_t in_pCount,
                             bool     in_visited_flg);

  static int InsertCardBoxInArray (const void *data, 
                                   const void *user);

  bool GetAllCardsInBox (int                   in_boxNum,
                         int                  *out_numDontNeedVisit_intP,
                         FlashCardInBoxType ***out_dontNeedVisit_cardsPAP,
                         int                  *out_numExpried_intP,
                         FlashCardInBoxType ***out_expired_cardsPAP);

  void MoveAllExpiredCards (uint32_t in_curTime_mins);
  void MoveToExpiredCards (CardBoxType *in_cardBoxP,
                           MoveModeEnum in_moveMode_enum,
                           uint32_t     in_curTime_mins);
  bool ResumeTimer ();
  void SaveTestSession (const char *in_testFile_str,
                        bool        in_isClosing_flg);
  void RestoreLastCardToUnprocessed ();

  char *m_cardBoxCabinetFile_str;
  char *m_cardBoxCabinetCmpPath_str;
  char *m_cardBoxCabinetDirOnly_str;

  CabinetXmlStateEnum m_cabinetXmlState_enum;
  int m_promoteCountThres;      // typically only for the first cardbox
  uint32_t m_oriBox1Study_mins; // duration in minutes for CardBox1
  uint32_t m_oriHigherBoxesReview_mins; // duration in minutes for higher boxes
  uint32_t m_box1Study_mins;         // actual used minutes for CardBox1
  uint32_t m_higherBoxesReview_mins; // actual used minutes for higher boxes

  
  bool            m_opened_flg;
  xmlListPtr      m_deckFileListP;
  FlashCardClass *m_flashCardObjA; // array of flash card objects.
  int             m_tmpFlashCardFileIdx;

  xmlListPtr    m_cardBoxesListP;
  CardBoxType  *m_tmpCardBoxP;
  int           m_tmpCardBoxIdx;
  int           m_numCardBoxes;
  int           m_neverVisitedCard_int;
  CardBoxType* *m_cardBoxPA;    // array of cardbox pointers
  int          *m_numDrawCards_intA; // size = m_numCardBoxes

  dict_t       *m_cbCabinet_dictP; // for tracking duplication and loading test 
  
  list_t   *m_studyDynSeqNode_listP; // for visiting prev cards
  lnode_t  *m_curStudyCardNode_lnodeP; // current study card: data = lnode

  StudySeqStateEnum   m_studySeqState_enum;
  lnode_t*           *m_studyCard_lnodePA; // Array of lnode pointers,
                                           // track current card in the
                                           // cardbox.
  int                 m_studyCardBox_idx;

  uint32_t m_startStudyTime_mins; // minutes since 1970
  uint32_t m_lastVisitedAt_mins; // minutes since 1970
  bool m_needSaving_flg;

  FlashCardInBoxType **m_dontNeedVisit_cardsPA;
  FlashCardInBoxType **m_expired_cardsPA;

  int  m_trashCardBox_idx;
  int *m_numExpireds_intA;
  int  m_reviewNthBoxOnlg_idx;
  int  m_lastCardBox_idx;       // the last cardbox that is not trash

  unsigned int m_puaseTimeout_uint;

  /* test-session support: */
  lnode_t            **m_test_lnodePA;
  uint8_t             *m_test_statusA; // TestStatusEnum
  int                  m_test_curIdx;
  int                  m_test_numCards;
  bool                 m_test_needSavingFlg;

  /* Restore last card to unprocessed: */
  uint32_t m_lastCardOriVisitedAt_mins;
  bool     m_reachedEndOfCards_flg;
};


/*--------------------*
 * Function Prototype *
 \*------------------*/
void* CreateStorage ();
void  DestroyStorage (void *storage);
void  ReAllocateStoreage (void *io_storagePP, // use void* instead of void**
                          int   in_bytes);


#endif // LEITNERSYSMODEL_HH
