/**
 * @file leitnerSysModelTester.cc
 *
 *  @remark ~ A testing program for the leitnersys library.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright (c) 2009 Tommy Chang. Permission is granted to copy,       .
 *  distribute and/or modify this document under the terms of the GNU    .
 *  Free Documentation License, Version 1.3 or any later version         .
 *  published by the Free Software Foundation; with no Invariant         .
 *  Sections, no Front-Cover Texts, and no Back-Cover Texts. A copy of   .
 *  the license is included in the file fdl-1.3.txt.                     .
 *  `--------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-06-07 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <getopt.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>


#include "leitnerSysModel.hh"

typedef struct {
  int         debugLevel_bitored;
  const char *statusFile_str;
} CommandLineParamType;



/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;
static CommandLineParamType g_clParams = {
  0,                            // debugLevel_bitored
  NULL,                         // statusFile_str
};
static int  g_quit_int = 0;
static bool g_executing_flag = false;




/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[leitnerSysModelTester %d]: ",     \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
// #ifdef NDEBUG
// #define fakeUseParam(...)
// #else
// static void fakeUseParam(...) {}
// #endif




static void
ExitFailure (const char *mesg_str = NULL)
{
  if (mesg_str)
    fprintf (stderr, "%s\n", mesg_str);

  if (g_clParams.statusFile_str)
    {
      FILE *sFileP = fopen (g_clParams.statusFile_str, "wb");
      fprintf (sFileP, "0"); // write 0 to status file -- failure
      fclose (sFileP);
    }
  exit (EXIT_FAILURE);
}


static void 
mysigint_handler (int)
{
  if (g_executing_flag)
    g_quit_int = 1;
  else
    {
      if (g_clParams.statusFile_str)
        {
          FILE *sFileP = fopen (g_clParams.statusFile_str, "wb");
          fprintf (sFileP, "0"); // write 0 to status file -- failure
          fclose (sFileP);
        }
      exit (EXIT_SUCCESS);
    }
}


// static double 
// mstime (void)
// {
//   double m;
//   struct timeval tp;
//   if (gettimeofday(&tp, 0) != 0)
//     return 0;
//   m = ((double) tp.tv_usec / 1000.0) + (tp.tv_sec * 1000);
//   return m;
// }


static void
CheckRequiredOptions ()
{


}


static void
SetSystemReady ()
{
  // system ready:
  if (g_clParams.statusFile_str)
    {
      FILE *sFileP = fopen (g_clParams.statusFile_str, "wb");
      fprintf (sFileP, "%d", getpid ()); // success = pid to status file 
      fclose (sFileP);
    }
}


static void
Run ()
{
  /*--------------------------------------
   * Check Require Commandline options:
   *------------------------------------*/
  CheckRequiredOptions ();

  /*-----------------------------
   * subsystems initialization:
   *---------------------------*/
  SetSystemReady ();

  /*-----------------------------
   * controller loop:
   *---------------------------*/
  fprintf (stderr, "calling %s\n", __FUNCTION__);
  g_executing_flag = true;
  FlashCardClass flashCardObj;
  flashCardObj.Load ("test_deck.dkf", NULL);
  flashCardObj.Save ("test_deck_save.dkf");
  fprintf (stderr, "sound path is '%s'\n", flashCardObj.GetSoundPath ());
  return;
  
  {
    LeitnerModelClass leitnerModelObj;
    for (int idx = 0; idx < 100000; idx ++)
      {
        if (leitnerModelObj.Open ("test.cbx") == false)
          assert (0);
        leitnerModelObj.LoadTestSession (NULL);
        leitnerModelObj.Close ();

        if (g_quit_int)
          break;
      }
  }
  return;


//--------------------------------------------------------
//--------------------------------------------------------
  flashCardObj.ChangeFields (0, 
                             "abase (hello)", // in_front_str
                             NULL, // in_back_str
                             NULL, // in_backExample_str
                             "tommy/chang", // in_frontPicFile_str
                             NULL); // in_backPicFile_str
  flashCardObj.Save ("hello.dkf");
  flashCardObj.Unload ();
  return;

//--------------------------------------------------------
//--------------------------------------------------------


  for (int idx = 0; idx < 100000; idx++)
    {
      flashCardObj.Load ("test_deck.dkf", NULL);
      fprintf (stderr, "sound path is '%s'\n", flashCardObj.GetSoundPath ());

      int numCards;
      const FlashCardType **flashCardsPA;
      flashCardsPA = flashCardObj.GetAllCards (&numCards);
      for (int idx = 0; idx < numCards; idx++)
        flashCardObj.PrintCardInfo (flashCardsPA [idx]);

      const FlashCardType *flashCardsP;
      flashCardsP = flashCardObj.GetCard (0);
      assert (flashCardsP);

      flashCardObj.Unload ();
      fprintf (stderr, "----------------------------------\n");
      if (g_quit_int)
        break;
    }
  return;

//--------------------------------------------------------
//--------------------------------------------------------

  LeitnerModelClass leitnerModelObj;
  {
    if (leitnerModelObj.Open ("cardBoxCabinet_test.xml") == false)
      assert (0);


    leitnerModelObj.GetNextCard ();
    leitnerModelObj.Promote ();
    leitnerModelObj.GetNextCard ();
    leitnerModelObj.Promote ();
    leitnerModelObj.GetNextCard ();
    leitnerModelObj.Promote ();

    leitnerModelObj.GetNextCard ();
    leitnerModelObj.Promote ();
    leitnerModelObj.GetNextCard ();
    leitnerModelObj.Promote ();

  }
}



static void
PrintUsage (char *progname_param)
{
  char *progname = strrchr (progname_param, '\\');
  if (progname == NULL)
    progname = progname_param;
  else
    progname += 1;

  fprintf 
    (stderr,
     "\n\nLast compiles on: %s\n"
     "Usage: %s [options]:\n"
     "[options] =\n"
     "  --help                Show this usage instruction.\n"
     "  --debug <int>         <int> is the level of verboseness;\n" 
     "                            can be multi-valued; default 0.\n" 
     "  --statusFile <str>    <str> is the output status file.\n" 
     "                            The status file contains the pid \n" 
     "                            if success, otherwise 0.\n" 
     "  --                    Allow no option supplied -- no effect.\n"
     "\n"
     "examples:\n"
     "  %s --\n"
     "  %s --debug 1 --debug 3\n"
     "\n", __DATE__
     , progname
     , progname
     , progname);
  exit (0);
}



static void
ParseCmdLine (int    argc,
              char **argv)
{
  /*---------------------------------
   * be intelligent about arguments
   *-------------------------------*/
  static struct option longOpt[] = {
    /*name, has_arg, int* flg, if flg: *flg = retval, ow. get_long() = retval*/
    {"debug",           1, NULL,        0},
    {"help",            0, NULL,       'h'},
    {"statusFile",      1, NULL,        0},
    {0,0,0,0} // terminator
  };


  if (argc == 1)
    PrintUsage (argv[0]);

  int opt = 0;
  int optIdx = 0;
# define IsOpt_M(pname) else if (strcmp (longOpt [optIdx].name, pname) == 0)
  while ((opt = getopt_long_only (argc, argv, "", longOpt, &optIdx)) != -1)
    {
      switch (opt)              // optarg is the argument
        {
        case 0:
          if (longOpt [optIdx].flag) // auto assigned value
            break;
          IsOpt_M ("debug")
            g_clParams.debugLevel_bitored |= (1 << (atoi (optarg) - 1));
          IsOpt_M ("statusFile")
            g_clParams.statusFile_str = optarg;
          else 
            fprintf (stderr, "un-handled option %s\n", longOpt [optIdx].name);
          break;
        case 'h':               // get help
          PrintUsage (argv[0]);
          exit (EXIT_SUCCESS);
          break;
        case '?':               // unrecognized option, no more msg needed
          ExitFailure ();
          break;
        default:
          printf ("%c option not implemented yet\n", opt);
          break;
        }
    }
# undef IsOpt_M

  if (optind < argc)
    {
      for (; optind < argc; optind++)
        fprintf (stderr, "unknown extra parameter %s\n", argv[optind]);
      ExitFailure ();
    }
}


int
main (int    argc,
      char **argv)
{
  /*------------------------------
   * parse command-line arguments:
   *----------------------------*/
  ParseCmdLine (argc, argv);


  /*-----------------------------
   * interrupt signals setup:
   *---------------------------*/
  signal (SIGINT, mysigint_handler);
  signal (SIGHUP, mysigint_handler);


  /*-----------------------------
   * run system:
   *---------------------------*/
  Run ();


  /*-----------------------------
   * program termination:
   *---------------------------*/
  fprintf (stderr, "Program Terminated Properly\n");
  return (EXIT_SUCCESS);
}
