/**
 * @file lookupTable.cc
 *
 *  @remark ~ Implementation of the generic look-up table support.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  . THIS software was developed at the National Institute of           .
 *  . Standards and Technology by employees of the Federal Government    .
 *  . in the course of their official duties. Pursuant to title 17       .
 *  . Section 105 of the United States Code this software is not         .
 *  . subject to copyright protection and is in the public               .
 *  . domain. NIST's [intelligent mobility] software is an experimental  .
 *  . system. NIST assumes no responsibility whatsoever for its use by   .
 *  . other parties, and makes no guarantees, expressed or implied,      .
 *  . about its quality, reliability, or any other characteristic. We    .
 *  . would appreciate acknowledgement if the software is used. This     .
 *  . software can be redistributed and/or modified freely provided      .
 *  . that any derivative works bear some notice that they are derived   .
 *  . from it, and any modified versions bear some notice that they      .
 *  . have been modified.                                                .
 *   `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2010-03-31 (Wed) 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include "lookupTable.hh"



/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;




/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[lookupTable %d]: ",               \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
// #ifdef NDEBUG
// #define fakeUseParam(...)
// #else
// static void fakeUseParam(...) {}
// #endif

LookUpTable::
LookUpTable () : m_table_hashP (NULL)
{
  m_table_hashP = hash_create (HASHCOUNT_T_MAX, 0, 0);
}

LookUpTable::
~LookUpTable ()
{
  // if (m_table_hashP == NULL)
  //   {
  //     fprintf (stderr, "nothing to free\n");
  //     return;
  //   }
  assert (m_table_hashP);

  /* destroy the hash table: */
  ResetLookUpTable ();
  hash_destroy (m_table_hashP);
  m_table_hashP = NULL;
}

bool LookUpTable::
SetLookupEntries (int         numKeys,
                  const char *keys_str, // multiple keys separated by '\0' 
                  const char *vals_str) // multiple values separated by '\0' 
{
  // note: keys_str could also be simulated by array of strings
  // keys_str[2][] = {"hello", "bye"}
  // and cast it to 1D, ie. (char *)
  // example:
  //   SetAttributes (2, "ctaTransRel_flg\0" "noMergeTag_flg",  "0\0" "0");
  fprintf (stderr, "calling SetLookupEntry");

  // if (unlikely (m_table_hashP == NULL))
  //   m_table_hashP = hash_create (HASHCOUNT_T_MAX, 0, 0);

  const char *key_str = keys_str;
  const char *val_str = vals_str;
  for (int idx = 0; idx < numKeys; idx++,
         key_str += strlen (key_str) + 1,
         val_str += strlen (val_str) + 1)
    {
      /* 
       * Update the hash:
       */
      /* is the key already exists? */
      hnode_t *hnodeP = hash_lookup (m_table_hashP, key_str);
      if (hnodeP == NULL)
        {
          /* no: need to create new entry: */
          if (hash_alloc_insert (m_table_hashP,
                                 strdup (key_str),
                                 strdup (val_str)) == 0)
	    {
	      fprintf (stderr, "could not insert key %s into hash\n", key_str);
	      return false;
	    }
        }
      else
        {
          /* yes: key already exist, replace the value: */
          assert (hnode_get (hnodeP));
          free (hnode_get (hnodeP));
          hnode_put (hnodeP, strdup (val_str));
        }
    }
  return true;
}


bool LookUpTable::
SetLookupEntry (const char *key_str,
                const void *val_voidP,
                int         valSize_int)
{
  // if (unlikely (m_table_hashP == NULL))
  //   m_table_hashP = hash_create (HASHCOUNT_T_MAX, 0, 0);

  /*
   * Duplicate the generic key: (old one will get freed)
   */
  uint8_t *val_int8A = (uint8_t*) malloc (valSize_int);
  memcpy (val_int8A, val_voidP, valSize_int);

  /* 
   * Update the hash:
   */
  /* is the key already exists? */
  hnode_t *hnodeP = hash_lookup (m_table_hashP, key_str);
  if (hnodeP == NULL)
    {
      /* no: need to create new entry: */
      if (hash_alloc_insert (m_table_hashP,
                             strdup (key_str),
                             val_int8A) == 0)
        {
          fprintf (stderr, "could not insert key into hash\n");
          return false;
        }
    }
  else
    {
      /* yes: key already exist, replace the value: */
      assert (hnode_get (hnodeP));
      free (hnode_get (hnodeP));
      hnode_put (hnodeP, val_int8A);
    }

  return true;                  // TBF
}



const void* LookUpTable::
GetLookUpEntry (const char *key_str)
{
  assert (m_table_hashP);

  /* check the hash table for key and value: */
  hnode_t *hnodeP = hash_lookup (m_table_hashP, key_str);
  if (hnodeP == NULL)
    {
      // fprintf (stderr, "Warning: unrecognized property key '%s'\n", key_str);
      // assert (0);
      return NULL;
    }
  
  return (const void*) hnode_get (hnodeP);
}


void LookUpTable::
ResetLookUpTable ()
{
  hscan_t  hs;
  hnode_t *hn;
  hash_scan_begin (&hs, m_table_hashP);
  while ((hn = hash_scan_next (&hs)))
    {
      char *val = (char*) hnode_get (hn);
      char *key = (char*) hnode_getkey (hn);
      hash_scan_delfree (m_table_hashP, hn);
      free (key);
      free (val);
    }
  assert (hash_count (m_table_hashP) == 0);
}


void LookUpTable::
Print (void (*printFun) (const char*, const void*))
{
  fprintf (stderr, "Table lookup entries are: \n");
  hscan_t  hs;
  hnode_t *hn;
  if (printFun == NULL)
    {
      hash_scan_begin (&hs, m_table_hashP);
      while ((hn = hash_scan_next (&hs)))
        {
          char *key = (char*) hnode_getkey (hn);
          char *val = (char*) hnode_get (hn);
          fprintf (stderr, "key = '%10s', val = '%10s'\n", key, val);
        }
      return;
    }

  hash_scan_begin (&hs, m_table_hashP);
  while ((hn = hash_scan_next (&hs)))
    printFun ((const char*) hnode_getkey (hn), hnode_get (hn));
}
