/*
 * List Abstract Data Type
 * Copyright (C) 1997 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: list.h,v 1.2 2007/04/22 17:41:15 tchang Exp $
 * $Name:  $  kazlib_1_20
 * 
 * Modified by Tommy Chang.
 */

#ifndef LIST_H
#define LIST_H

#include <limits.h>

/*
 * Blurb for inclusion into C++ translation units
 */

#ifdef __cplusplus
extern "C" {
#endif
  

  /*----------------------
   * Node Data structure :
   *---------------------*/
  typedef unsigned long listcount_t;
  
  typedef struct lnode_t {
    struct lnode_t *list_next;
    struct lnode_t *list_prev;
    void *list_data;
  } lnode_t;

  typedef struct lnodepool_t {
    struct lnode_t *list_pool;
    struct lnode_t *list_free;
    listcount_t list_size;
  } lnodepool_t;

#define LISTCOUNT_T_MAX ULONG_MAX


  /*----------------------
   * List Data structure :
   *---------------------*/
  typedef struct list_t {
    lnode_t list_nilnode;
    listcount_t list_nodecount;
    listcount_t list_maxcount;
  } list_t;



  /*----------------------
   * Node Operations :
   *---------------------*/
  lnode_t* lnode_create       (void*);
  lnode_t* lnode_init         (lnode_t*, void*);
  void     lnode_destroy      (lnode_t*);
  int      lnode_is_in_a_list (lnode_t*);


  lnodepool_t* lnode_pool_init    (lnodepool_t*, lnode_t*, listcount_t);
  lnodepool_t* lnode_pool_create  (listcount_t);
  void         lnode_pool_destroy (lnodepool_t*);
  lnode_t*     lnode_borrow       (lnodepool_t*, void*);
  void         lnode_return       (lnodepool_t*, lnode_t*);
  int          lnode_pool_isfrom  (lnodepool_t*, lnode_t*);

#define lnode_put(N, D)         ((N)->list_data = (D))
#define lnode_get(N)            ((N)->list_data)
#define lnode_pool_isempty(P)   ((P)->list_free == 0)


  /*----------------------
   * LIST Operations :
   *---------------------*/
  list_t*  list_init          (list_t*, listcount_t);
  list_t*  list_create        (listcount_t);
  void     list_destroy       (list_t*);
  void     list_destroy_nodes (list_t*);
  void     list_return_nodes  (list_t*, lnodepool_t*);
  int      list_contains      (list_t*, lnode_t*);
  void     list_ins_before    (list_t*, lnode_t*, lnode_t*);
  void     list_ins_after     (list_t*, lnode_t*, lnode_t*);
  lnode_t* list_delete        (list_t*, lnode_t*);
  int      list_verify        (list_t *);
  void     list_process       (list_t *, void *, 
                               void (*)(list_t*, lnode_t*, void*));

  /* destination list on the left, source on the right */
  void     list_extract   (list_t*, list_t*, lnode_t*, lnode_t*);
  void     list_transfer  (list_t*, list_t*, lnode_t*);
  void     list_merge     (list_t*, list_t*, int (const void*, const void*));
  void     list_sort      (list_t*, int (const void*, const void*));
  lnode_t* list_find      (list_t*, const void*, 
                           int (const void*, const void*));
  int      list_is_sorted (list_t*, int (const void *, const void *));

  /* TC 2008-12-13 -- Section below Added by Tommy Chang  */
  void     list_free_and_destroy (list_t*);

  /* TC 2005-06-23 -- Section below Added by Tommy Chang  */
  void     list_return_nodes_fast (list_t*, lnodepool_t*);
  void     list_transfer_fast     (list_t*, list_t*);
  void*    fastLnode_pool_create  (int); 
  void     fastLnode_pool_destroy (void*);
  lnode_t* fastLnode_pool_borrow  (void*);
  void     fastLnode_pool_collect (void*);


#define list_count(L)           ((L)->list_nodecount)
#define list_isempty(L)         ((L)->list_nodecount == 0)
#define list_isfull(L)          ((L)->list_nodecount == (L)->list_maxcount)
#define list_next(L, N)         ((N)->list_next == &(L)->list_nilnode ? \
                                 NULL : (N)->list_next)
#define list_prev(L, N)         ((N)->list_prev == &(L)->list_nilnode ? \
                                 NULL : (N)->list_prev)
#define list_first(L)           list_next(L, &(L)->list_nilnode)
#define list_last(L)            list_prev(L, &(L)->list_nilnode)
#define list_nil(L)		(&(L)->list_nilnode)

#define list_append(L, N)       list_ins_before(L, N, &(L)->list_nilnode)
#define list_prepend(L, N)      list_ins_after(L, N, &(L)->list_nilnode)
#define list_del_first(L)       list_delete(L, list_first(L))
#define list_del_last(L)        list_delete(L, list_last(L))

#ifdef __cplusplus
}
#endif

#endif
